(* as_sig.v *)
(* Specifications of the "as" module over the PMAP machine model. *)

Require Import List.
Require Import Bool.
Require Import NArith.
Require Import ZArith.

Require Import spec.
Require Import label.

Require Import zbool_assist.
Require Import base_simp.
Require Import Omega.
Require Import Program.

Require Import pmap.
Require Import pmap_lib.
Require Import c.

Require Import sact.
Require Import actsmp.

Module as_sig.
Import C_PMAP SCAP.

Notation "a ::: b" := (Iseq a b) (at level 80).
Notation "a ;; b" := (ActionChain a b) (at level 80).

Definition PMAP_valloc (vpg : Z) (pg : Z) : PMAP_State.t -> PMAP_State.t -> Prop :=
  fun s s' => 
         (forall l, s.(allocdata) (Pg l) = true -> s'.(data) l = s.(data) l) /\
	 (s'.(allocdata) = (fun x => if Zeq_bool x pg then true else s.(allocdata) x)) /\
	(s'.(pagemap) = (fun v => if Zeq_bool v vpg then Some pg else s.(pagemap) v)).

Definition vallocpg (vpg pg : Z) : State -> State -> Prop :=
  fun s s' => PMAP_valloc vpg pg s.(memory) s'.(memory) /\ StackRets vpg s.(stack) s'.(stack).


Lemma PMAP_valloc_trans_unch :  forall vp pg l s s',
Pg l <> vp ->
as_sig.PMAP_valloc vp pg s s' ->
PMAP_State.trans l s = PMAP_State.trans l s'.
intros.
red in H0; decompose record H0; clear H0.
unfold PMAP_State.trans.
rewrite H4.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Qed.

Lemma PMAP_valloc_trans'_unch :  forall vp pg l s s',
Pg l <> vp ->
as_sig.PMAP_valloc vp pg s s' ->
PMAP_State.trans' l s = PMAP_State.trans' l s'.
intros.
unfold PMAP_State.trans'.
makenew' (isPhys l); autosolve.
apply PMAP_valloc_trans_unch with vp pg; auto.
Qed.

Lemma PMAP_valloc_allocdata_unch : forall vp pg pg' s s',
pg' <> pg ->
as_sig.PMAP_valloc vp pg s s' ->
PMAP_State.allocdata s' pg' = PMAP_State.allocdata s pg'.
intros.
red in H0; decompose record H0; clear H0.
rewrite H3; simpl.
Zeq_pull'; Zeq_norm; simpl; autosolve.
Qed.

Lemma PMAP_valloc_pagemap_unch : forall vp pg vp' s s',
vp' <> vp ->
as_sig.PMAP_valloc vp pg s s' ->
PMAP_State.pagemap s' vp' = PMAP_State.pagemap s vp' .
intros.
red in H0; decompose record H0; clear H0; autosolve.
rewrite H4.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Qed.

Lemma PMAP_valloc_data_unch : forall l vp x0 s s',
PMAP_State.allocdata s (Pg l) = true ->
as_sig.PMAP_valloc vp x0 s s' ->
PMAP_State.data s' l = PMAP_State.data s l.
intros.
red in H0; decompose record H0; clear H0.
rewrite H1; autosolve.
Qed.


Definition as_reserve_spec : Action State := {|
pre := (fun s => StackCheck (getArgsCheck 1) s = true /\
                                  let vpage := nth 0 (StackGet getArgs s) 0%Z in
                                  ValidVPg vpage = true /\ ValidPg vpage = false 
                                  );
rel := (fun s s' =>
       let vpage := nth 0 (StackGet getArgs s) 0%Z in
       (exists pg , (1 <= pg < NPAGES)%Z /\ s.(memory).(allocdata) pg = false /\ vallocpg vpage pg s s')  
       \/ 
       (s' = StackAct (pop +++ setRet 0%Z) s))
|}.






Definition PMAP_vfree (vpg : Z) : PMAP_State.t -> PMAP_State.t -> Prop :=
  fun s s' => 
        (forall l, s'.(allocdata) (Pg l) = true -> s'.(data) l = s.(data) l) /\
	(s'.(pagemap) = (fun v => if Zeq_bool v vpg then None else s.(pagemap) v)) /\
        match s.(pagemap) vpg with
	| Some pg => 
		  (s'.(allocdata) = (fun x => if Zeq_bool x pg then false else s.(allocdata) x))
	| None => False
	end.

Definition vfree (vpg: Z) : State -> State -> Prop :=
  fun s s' => PMAP_vfree vpg s.(memory) s'.(memory) /\ StackRets 0%Z s.(stack) s'.(stack).




Definition as_release_spec : Action State := {|
pre := (fun s => StackCheck (getArgsCheck 1) s = true /\
                                  let vpage := nth 0 (StackGet getArgs s) 0%Z in
                                  ValidVPg vpage = true /\ ValidPg vpage = false /\ 
                                  match s.(memory).(pagemap) vpage with
                                     | Some ppg => ValidPg ppg = true /\ ppg <> 0%Z /\ s.(memory).(allocdata) ppg = true
                                     | None => False
                                  end
                                  );
rel := (fun s s' =>
       let vpage := nth 0 (StackGet getArgs s) 0%Z in
       vfree vpage s s'
	)
|}.

Definition as_spec : SpecHeap := fun l =>
  match l with
  | as_reserve_label => Some as_reserve_spec
  | as_release_label => Some as_release_spec
  | _ => None
  end.

Lemma lib_as_disj : C_PMAP.SCAP.SpecHeap.SpecHeap_disjoint PMAP_Lib.pmap_lib_spec as_sig.as_spec.
red; intros.
destruct l; autosolve.
Qed.

Definition AS_kernel_init_spec : Action C_AS.State.t := {|
pre := (fun s => C_AS.State.StackCheck (C_AS.State.Stack.getArgsCheck 0) s = true);
rel := (fun s s' => False)
|}.

Definition PMAP_kernel_init_spec : Action C_PMAP.State.t := {|
pre := (fun s => 
     C_PMAP.State.StackCheck (C_PMAP.State.Stack.getArgsCheck 0) s = true /\
     (forall pg, ValidVPg pg = true -> ValidPg pg = false -> s.(C_PMAP.State.memory).(PMAP_State.pagemap) pg = None)
      );
rel := (fun s s' => False)
|}.


End as_sig.

