(* c.v  - An implementation of the C machine *)
(* Defines a general C machine, its semantics, and verification system *)

Require Import Equalities.
Require Import DecidableType.
Require Import ZArith.
Require Import NArith.
Require Import Bool.
Require Import IfProp.

Require Import spec.
Require Import scap.
Require Import zbool_assist.

Require Import Logic.FunctionalExtensionality.

Create HintDb Zbool_simpl discriminated.
Hint Rewrite Zeq_eq_bool using assumption : Zbool_simpl.
Hint Rewrite Zneq_eq_bool_false using assumption : Zbool_simpl.
Hint Rewrite Zeq_bool_refl : Zbool_simpl.

Require Export common.
Require Import sact.
Require Import actsmp.
Require Import mem.


(* FLABEL is not very useful right now. It is needed for future work *)
Definition FLABEL := Z.

(*********************************************)
(* The type of Variables *)
Definition Var := Z.
Definition Var_eq := @eq Z.
Definition Var_beq := Zeq_bool.
Definition Var_eq_dec := Z_eq_dec.

(*********************************************)
(* The definition of Stack in the C machine *)
Module C_Stack <: Typ.
Definition dataframe := Var -> option Z.

Inductive Frame : Type :=
  | DataFrame : dataframe -> Frame
  | CallFrame : list Z -> Frame
  | RetFrame : Z -> Frame.

Definition t := list Frame.

Definition getRetCheck : SCheck (State := t) :=
fun s => match s with
  | RetFrame _ :: _ => true
  | _ => false
end.
Definition getRet : SGet (State:=t) Z := 
fun s => match s with
  | RetFrame z :: _ => z
  | _ => 0%Z
end.

Definition getVarCheck (v : Var) : SCheck (State := t) :=
fun s =>
  match s with
   | DataFrame vars :: _ => match vars v with | Some _ => true  | _ => false end
   | _ => false
  end.

Definition getVar (v : Var) : SGet (State :=t) Z := 
fun s =>
  match s with
  | DataFrame vars :: _ => match vars v with | Some z => z | None => 0%Z end 
  | _ => 0%Z
  end.

Definition getArgsCheck (n : nat) : SCheck (State := t) :=
fun s =>
  match s with
  | CallFrame al :: _ => beq_nat n (length al)
  | _ => false
  end.
Definition getArgs : SGet (State := t) (list Z) :=
fun s =>
  match s with
  | CallFrame al :: _ => al
  | _ => nil
  end.

Definition isDataFrame : SCheck(State := t) := 
fun s => match s with
           | DataFrame _ :: _ => true
           | _ => false
   end.
Definition isFrame : SCheck (State := t) :=
fun s =>  match s with
           | _ :: _ => true
           | _ => false
   end.

Definition setRetCheck (z : Z) : SCheck (State := t) := fun s => true.
Definition setRet (z : Z) : SAct (State := t) := 
fun s => RetFrame z :: s.

Definition setVarCheck (z : Z) : SCheck (State := t) := 
  fun s => 
        match s with
        | DataFrame _ :: _ => true
        | _ => false
        end.
Definition setVar (v : Var) (z : Z) : SAct (State := t) := 
  fun s => 
         match s with
         | DataFrame vars :: stack' => 
            DataFrame (fun v' => if Var_beq v' v then Some z else vars v') :: stack'
         | _ => s
         end.

Definition pushVarCheck : SCheck (State := t) := fun s => true.
Definition pushVar : SAct (State := t) :=
  fun s => DataFrame (fun v' => None) :: s.

Definition pushArgsCheck : SCheck (State := t) := fun s => true.
Definition pushArgs (al : list Z) : SAct (State := t) := 
  fun s => CallFrame al :: s.

Definition popCheck : SCheck (State := t) := 
  fun s => 
    match s with 
     | _ :: stack' => true
     | _ => false
    end.
Definition pop : SAct (State := t) := 
  fun s => 
    match s with
      | _ :: stack' => stack'
      | _ => s
    end.


(* getVar simplifications *)
Lemma getVar_setRet : forall v z s, getVar v (setRet z s) = 0%Z.
firstorder. Qed.

Lemma getVar_setVar : forall v v' z' s, getVar v (setVar v' z' s) = 
   if isDataFrame s then 
        (if Var_beq v v' then z' else getVar v s)
   else 0%Z.
intros.
unfold getVar, setVar, Var_beq; firstorder.
destruct s; simpl; auto.
destruct f; simpl; auto.
makenew' (Zeq_bool v v'); auto.
Qed.

Lemma getVar_pushVar : forall v s, getVar v (pushVar s) = 0%Z.
destruct s; simpl; auto. Qed.

Lemma getVar_pushArgs : forall v l s, getVar v (pushArgs l s) = 0%Z.
destruct s; simpl; auto. Qed.

Hint Rewrite getVar_setRet getVar_setVar getVar_pushVar getVar_pushArgs : stack.

Lemma getVarCheck_setRet : forall v z s, getVarCheck v (setRet z s) = false.
firstorder. Qed.

Lemma getVarCheck_pushVar : forall v s, getVarCheck v (pushVar s) = false.
firstorder. Qed.

Lemma getVarCheck_pushArgs : forall v al s, getVarCheck v (pushArgs al s) = false.
firstorder. Qed.

Lemma getVarCheck_setVar : forall v v' z' s, getVarCheck v (setVar v' z' s) = 
   if Var_beq v v' then isDataFrame s else getVarCheck v s.
firstorder.
unfold getVarCheck, setVar, isDataFrame, Var_beq.
destruct s; try destruct f; simpl; autosolve; Zeq_pull'; Zeq_norm; autosolve.
Qed.

Hint Rewrite getVarCheck_setRet getVarCheck_pushVar getVarCheck_pushArgs getVarCheck_setVar : stack.

Lemma getRet_setRet : forall z s, getRet (setRet z s) = z.
auto.
Qed.

Hint Rewrite getRet_setRet : stack.

(* isDataFrame simplifications *)
Lemma isDataFrame_setRet : forall z s, isDataFrame (setRet z s) = false.
destruct s; simpl; auto. Qed.
Lemma isDataFrame_setVar : forall v z s, isDataFrame (setVar v z s) = isDataFrame s.
destruct s; simpl; auto.
destruct f; simpl; auto.
Qed.
Lemma isDataFrame_pushVar : forall s, isDataFrame (pushVar s) = true.
destruct s; simpl; auto. Qed.
Lemma isDataFrame_pushArgs : forall l s, isDataFrame (pushArgs l s) = false.
destruct s; simpl; auto. Qed.

Lemma isDataFrame_isFrame : forall s, isDataFrame s = true -> isFrame s = true.
intros.
unfold isDataFrame, isFrame; destruct s; simpl in *; auto.
Qed.

Lemma getArgsCheck_isFrame : forall n s, getArgsCheck n s = true -> isFrame s = true.
intros; unfold getArgsCheck, isFrame in *.
destruct s; simpl in *; auto.
Qed.

Hint Rewrite isDataFrame_setRet isDataFrame_setVar isDataFrame_pushVar 
             isDataFrame_pushArgs : stack.
Hint Resolve isDataFrame_isFrame getArgsCheck_isFrame.


(* isFrame simplifications *)
Lemma isFrame_setRet : forall z s, isFrame (setRet z s) = true.
destruct s; simpl; auto. Qed.
Lemma isFrame_setVar : forall v z s, isFrame (setVar v z s) = isFrame s.
destruct s; simpl; auto.
destruct f; simpl; auto.
Qed.

Lemma isFrame_pushVar : forall s, isFrame (pushVar s) = true.
destruct s; simpl; auto. Qed.
Lemma isFrame_pushArgs : forall l s, isFrame (pushArgs l s) = true.
destruct s; simpl; auto. Qed.

Hint Rewrite isFrame_setRet isFrame_setVar isFrame_pushVar 
             isFrame_pushArgs : stack.


(* pop simplification *)
Lemma pop_setRet : forall r s, pop (setRet r s) = s.
destruct s; simpl; auto.
Qed.

Lemma pop_setVar : forall v z s, pop (setVar v z s) = pop s.
unfold pop, setVar. 
destruct s; simpl; auto.
makenew' f; simpl; auto.
Qed.

Lemma pop_pushVar : forall s, pop (pushVar s) = s.
destruct s; simpl; auto.
Qed.

Lemma pop_pushArgs : forall l s, pop (pushArgs l s) = s.
destruct s; simpl; auto.
Qed.

Hint Rewrite pop_setRet pop_setVar pop_pushVar pop_pushArgs : stack.

Lemma setVar_unneeded : forall v z s, getVarCheck v s = true -> (getVar v s = z) -> (setVar v z s = s).
destruct s; simpl; auto.
makenew' f; autosolve.
unfold getVar, setVar, getVarCheck ; simpl.
intros.
repeat f_equal.
apply functional_extensionality; intros.
unfold Var_beq.
Zeq_pull'; Zeq_norm; autosolve.
subst; autosolve.
makenew' (d v); simpl; autosolve.
Qed.

Hint Resolve getArgsCheck_isFrame : stack.

Lemma getRetCheck_setRet : forall x s, getRetCheck (setRet x s) = true.
intros.
unfold getRetCheck, setRet; auto.
Qed.

Hint Rewrite getRetCheck_setRet : stack.


Lemma getVarCheck_isDataFrame : forall v s, getVarCheck v s = true -> (isDataFrame s = true).
destruct s; simpl; autosolve.
destruct f; autosolve.
Qed.

Lemma setVar_order : forall v z v' z' s, Var_beq v v' = false ->
   setVar v z (setVar v' z' s) = setVar v' z' (setVar v z s).
destruct s; simpl; autosolve; intros.
destruct f; simpl; autosolve.
unfold setVar; simpl.
repeat f_equal; autosolve.
apply functional_extensionality; intros.
unfold Var_beq in *.
Zeq_pull'; Zeq_norm; autosolve.
Qed.

Lemma setVar_dbl : forall v z v' z' s, Var_beq v v' = true ->
  setVar v z (setVar v' z' s) = setVar v z s.
destruct s; simpl; autosolve; intros.
destruct f; simpl; autosolve.
unfold setVar; simpl.
repeat f_equal.
apply functional_extensionality; intros.
unfold Var_beq in *.
Zeq_pull'; Zeq_norm; autosolve.
Qed.

Lemma setVar_dbl2 : forall v z z' s, setVar v z (setVar v z' s) = setVar v z s.
destruct s; simpl; autosolve; intros.
destruct f; autosolve; simpl; intros.
unfold setVar; simpl.
repeat f_equal.
apply functional_extensionality; intros.
unfold Var_beq in *.
Zeq_pull'; Zeq_norm; autosolve.
Qed.
Hint Rewrite setVar_dbl2 : stack.

Fixpoint setVarsCheck' (vl : list Var) (zl : list Z) : SCheck (State := t) :=
  match (vl, zl) with
  | (v :: vl', z :: zl') => setVarsCheck' vl' zl' 
  | (nil, nil) => fun s => true
  | _ => fun s => false
  end.
Definition setVarsCheck (vl : list Var) (zl : list Z) : SCheck := 
  SCheck_and isDataFrame (setVarsCheck' vl zl). 

Fixpoint setVars (vl : list Var) (n : nat) (zl : list Z) : SAct :=
  match vl with
  | v :: vl' => setVar v (nth n zl 0%Z) +++ setVars vl' (S n) zl
  | nil => fun s => s
  end.

Definition loadArgsCheck (lv : list Var) : SCheck := getArgsCheck (length lv).
 
Definition loadArgs (lv : list Var) : SAct :=
SAct_Let getArgs (fun la => pop +++ pushVar +++ (setVars lv 0 la)). 



Definition StackRets (ret : Z) : t -> t -> Prop :=
  fun s s' => s' = (pop +++ setRet ret%Z) s.

End C_Stack.

(************************************************************************)
(* Expressions *)
Inductive UNOP : Set := | BANG | UMINUS.
Inductive BINOP : Set := 
| PLUS | MINUS | TIMES | DIV | MOD | LOGIC_AND | LOGIC_OR | EQ | NEQ
| LT | GT | LE | GE.

Scheme Equality for UNOP.
Scheme Equality for BINOP.

Definition bop_op (bop : BINOP) : Z -> Z -> Z :=
  match bop with
  | PLUS => Zplus
  | MINUS => Zminus
  | TIMES => Zmult
  | DIV => Zdiv
  | MOD => Zmod
  | LOGIC_AND => Zmult
  | LOGIC_OR => Zplus
  | EQ => fun a b => if Zeq_bool a b then 1%Z else 0%Z
  | NEQ => fun a b => if Zeq_bool a b then 0%Z else 1%Z
  | LT => fun a b => if Zlt_bool a b then 1%Z else 0%Z
  | GT => fun a b => if Zgt_bool a b then 1%Z else 0%Z
  | LE => fun a b => if Zle_bool a b then 1%Z else 0%Z
  | GE => fun a b => if Zge_bool a b then 1%Z else 0%Z
  end.

Definition option_binop_lift (A B C : Type) (bop : A -> B -> C) 
  : (option A -> option B -> option C) :=
  fun oa ob =>
  match (oa, ob) with
  | (Some a, Some b) => Some (bop a b)
  | _ => None
  end.
Implicit Arguments option_binop_lift [A B C].

Definition option_unop_lift (A B : Type) (uop : A -> B) : (option A -> option B) :=
  fun oa => match oa with
  | Some a => Some (uop a)
  | None => None
  end.
Implicit Arguments option_unop_lift [A B].


Definition eval_bop_opt (bop : BINOP) : (option Z -> option Z -> option Z) :=
  option_binop_lift (bop_op bop).

Definition uop_op (uop : UNOP) : Z -> Z :=
  match uop with
  | BANG => fun a => if Zeq_bool a 0 then 1%Z else 0%Z
  | UMINUS => Zminus 0
  end.

Definition eval_uop (uop : UNOP) : (option Z -> option Z) :=
  option_unop_lift (uop_op uop).

(**********************************************************************)
(* Expressions *)

Module Expr <: Typ.

(* Expressions that are used in conditionals  - as they do not access memory *)
Inductive expr_nomem : Set :=
  | expr_nomem_z : Z -> expr_nomem
  | expr_nomem_var : Z -> expr_nomem
  | expr_nomem_binop : BINOP -> expr_nomem -> expr_nomem -> expr_nomem
  | expr_nomem_unop : UNOP -> expr_nomem -> expr_nomem.

(* Full expression *)
Inductive expr : Set := 
  | expr_z : Z -> expr
  | expr_var : Z -> expr
  | expr_mem : expr -> expr
  | expr_binop : BINOP -> expr -> expr -> expr
  | expr_unop : UNOP -> expr -> expr.

Fixpoint expr_nomem_embed (e : expr_nomem) : expr :=
match e with
   | expr_nomem_z z => expr_z z
   | expr_nomem_var v => expr_var v
   | expr_nomem_binop b e1 e2 => expr_binop b (expr_nomem_embed e1) (expr_nomem_embed e2)
   | expr_nomem_unop u e1 => expr_unop u (expr_nomem_embed e1)
end.

Definition t := expr_nomem.
End Expr.

Definition expr := Expr.expr.
Import Expr.
Export Expr.


(***********************************************************)
(* Definition of the C machine State *)

Module C_State (Mem : Mem_T) <: Typ. (* DecidableType. *)
Module Stack := C_Stack.
Module Mem := Mem.
Export Stack.
Export Mem.


Record State := {stack : C_Stack.t;  memory : Mem.t}.

Definition t := State.

Definition StackCheck (f : SCheck (State := C_Stack.t)) : SCheck (State := t) :=
  fun s => (f (s.(stack))).

Definition StackGet {A} (f : SGet (State := C_Stack.t) A) : SGet (State := t) A :=
  fun s => f (s.(stack)).

Definition StackAct (f : SAct (State := C_Stack.t)) : SAct (State := t) :=
  fun s => {| stack := f (s.(stack)); memory := s.(memory) |}.

Definition MemCheck (f : SCheck (State := Mem.t)) : SCheck (State := t) :=
  fun s => (f (s.(memory))).

Definition MemGet {A} (f : SGet (State := Mem.t) A) : SGet (State := t) A :=
  fun s => f (s.(memory)).

Definition MemAct (f : SAct (State := Mem.t)) : SAct (State := t) :=
  fun s => {| stack := s.(stack); memory := f (s.(memory)) |}.

(* Order of actions *)
Lemma Stack_Mem_Act : forall st m s, MemAct m (StackAct st s) = StackAct st (MemAct m s).
firstorder. Qed.

Hint Rewrite Stack_Mem_Act (* StackAct_StackAct MemAct_MemAct *) : c.

(* Stack get operations *)
Lemma StackGet_MemAct : forall A (get : SGet (State := C_Stack.t) A) m s, StackGet get (MemAct m s) = StackGet get s.
firstorder. Qed.

Lemma StackCheck_MemAct : forall (ch : SCheck (State := C_Stack.t)) m s, StackGet ch (MemAct m s) = StackCheck ch s.
firstorder. Qed.

Lemma StackGet_StackAct : forall A (get : SGet (State := C_Stack.t) A) m s, StackGet get (StackAct m s) = StackGet (fun s' => get (m s')) s.
firstorder. Qed.

Lemma StackCheck_StackAct : forall (ch : SCheck (State := C_Stack.t)) m s, StackGet ch (StackAct m s) = StackCheck (fun s' => ch (m s')) s.
firstorder. Qed.


(* Memory get operations *)
Lemma MemGet_StackAct : forall A (get : SGet (State := Mem.t) A) m s, MemGet get (StackAct m s) = MemGet get s.
firstorder. Qed.

Lemma MemCheck_StackAct : forall (ch : SCheck (State := Mem.t)) m s, MemGet ch (StackAct m s) = MemCheck ch s.
firstorder. Qed.

Lemma MemGet_MemAct : forall A (get : SGet (State := Mem.t) A) m s, 
                               MemGet get (MemAct m s) = MemGet (fun mem => get (m mem)) s.
firstorder. Qed.

Lemma MemCheck_MemAct : forall (ch : SCheck (State := Mem.t)) m s, MemGet ch (MemAct m s) = MemCheck (fun mem => ch (m mem)) s.
firstorder. Qed.

Hint Rewrite StackGet_MemAct StackCheck_MemAct StackGet_StackAct StackCheck_StackAct : c.
Hint Rewrite MemGet_MemAct MemCheck_MemAct MemGet_StackAct MemCheck_StackAct : c.

Fixpoint eval(e : expr) : SGet Z := fun s =>
  match e with
  | expr_z z => z
  | expr_var v => StackGet (getVar v) s
  | expr_mem e => MemGet (getMem (eval e s)) s
  | expr_binop bop e1 e2 => (bop_op bop) (eval e1 s) (eval e2 s)
  | expr_unop uop e => (uop_op uop) (eval e s)
  end.


Fixpoint evalCheck (e : expr) : SCheck (State := t) :=
  match e with
  | expr_z z => (fun s => true)
  | expr_var v => StackCheck (getVarCheck v)
  | expr_mem e =>  SCheck_and (
                                       fun s =>
                                          MemCheck (getMemCheck (eval e s)) s)
                               (evalCheck e)
  | expr_binop bop e1 e2 => SCheck_and (evalCheck e1) (evalCheck e2)
  | expr_unop uop e => evalCheck e
  end.


Fixpoint evallistCheck (el : list expr) : SCheck :=
  match el with
  | nil => (fun s => true)
  | e :: el' => SCheck_and (evalCheck e) (evallistCheck el')
  end.

Fixpoint evallist' (el : list expr) (s : State) : list Z :=
  match el with
  | nil => nil
  | e :: el' => (eval e s :: evallist' el' s) 
  end.
Definition evallist (el : list expr) : SGet (list Z) := evallist' el.

Ltac rewriter := 
match goal with
 | h1 : Some ?a = Some ?b |- _ => inversion h1; clear h1; subst
 | h1 : ?a = Some ?b |- _ => 
    match type of a with
    | option State => symmetry in h1
    | _ => fail
    end
 | h1 : Some ?a = ?b |- _ => rewrite h1 in * |- *; try clear a h1
 | h1 : ?a = ?b |- _ => rewrite h1 in * |- *
end.

Definition expr_to_Prop (e : expr_nomem) := 
   fun s => (evalCheck (Expr.expr_nomem_embed e) s = true ) /\ 
                (eval (Expr.expr_nomem_embed e) s <> 0)%Z.

End C_State. 




(****************************************************************)
(* C machine operations *)



Module C_Ops <: Typ.

Inductive perf : Set :=
  | perf_assign : Var -> expr -> perf
  | perf_fcall : FLABEL -> list expr -> perf
  | perf_ret : expr -> perf
  | perf_readargs : list Var -> perf
  | perf_readret : Var -> perf
  | perf_store : expr -> expr -> perf
.

 Definition t := perf.
End C_Ops.

(********************************************************************)
(* The complete C machine, with verification system *)
(* We have to define it as a module type and as a module so that the definitions in every
    instance of the C machine remain the same *)


Module Type C_T (Mem : Mem_T).
Module State := C_State Mem.
Module Ops := C_Ops.
Export State.
Import State.Mem.
Import State.Stack.
Export Ops.

Module SCAP := SCAP State Ops Expr. 
Export SCAP.
Import SCAP.
Import SCAP.Process.

Definition Actions (i : perf) : Action State :=
  match i with
  | perf_assign v e => SAct_Action (SCheck_and (evalCheck e)  (StackCheck (setVarCheck v)))
                                                           (SAct_Let (eval e) (fun z => StackAct (setVar v z)))
  | perf_fcall f el => SAct_Action (evallistCheck el) (SAct_Let (evallist el) (fun z => StackAct (pushArgs z)))
  | perf_readargs vl => SAct_Action (StackCheck (loadArgsCheck vl)) (StackAct (loadArgs vl))
  | perf_readret v => SAct_Action (StackCheck getRetCheck)
                                        (SAct_Let (StackGet getRet) (fun z => StackAct (pop +++ (setVar v z))))
  | perf_store eloc e => SAct_Action 
                         (SCheck_and (StackCheck isDataFrame) 
                              (SCheck_and (evalCheck eloc)
				(SCheck_and (evalCheck e)
                            	  (SCheck_Let (eval eloc) (fun va => MemCheck (setMemCheck va))))))

                         (SAct_Let (eval eloc) (fun va =>
                         SAct_Let (eval e) (fun z =>
                         (MemAct (setMem va z))
                         )))
  | perf_ret e => SAct_Action (SCheck_and (StackCheck isFrame) (evalCheck e)) 
                           (SAct_Let (eval e) (fun z => StackAct (pop +++ (setRet z))))
  end.
 
Definition Mc_Machine := fun c => Some (Actions c).

End C_T.

Module C (Mem : Mem_T) <: (C_T Mem).
Module State := C_State Mem.
Module Ops := C_Ops.
Export State.
Import State.Mem.
Import State.Stack.
Export Ops.

Module SCAP := SCAP State Ops Expr.
Export SCAP.
Import SCAP.
Import SCAP.Process.

Definition Actions (i : perf) : Action State :=
  match i with
  | perf_assign v e => SAct_Action (SCheck_and (evalCheck e)  (StackCheck (setVarCheck v)))
                                                           (SAct_Let (eval e) (fun z => StackAct (setVar v z)))
  | perf_fcall f el => SAct_Action (evallistCheck el) (SAct_Let (evallist el) (fun z => StackAct (pushArgs z)))
  | perf_readargs vl => SAct_Action (StackCheck (loadArgsCheck vl)) (StackAct (loadArgs vl))
  | perf_readret v => SAct_Action (StackCheck getRetCheck)
                                        (SAct_Let (StackGet getRet) (fun z => StackAct (pop +++ (setVar v z))))
  | perf_store eloc e => SAct_Action 
                         (SCheck_and (StackCheck isDataFrame) 
                              (SCheck_and (evalCheck eloc)
				(SCheck_and (evalCheck e)
                            	  (SCheck_Let (eval eloc) (fun va => MemCheck (setMemCheck va))))))

                         (SAct_Let (eval eloc) (fun va =>
                         SAct_Let (eval e) (fun z =>
                         (MemAct (setMem va z))
                         )))
  | perf_ret e => SAct_Action (SCheck_and (StackCheck isFrame) (evalCheck e)) 
                           (SAct_Let (eval e) (fun z => StackAct (pop +++ (setRet z))))
  end.
 
Definition Mc_Machine := fun c => Some (Actions c).

(* autounfold all confusing C details *)
Ltac C_auf :=
repeat try (
unfold Actions in *;
unfold loadArgs, loadArgsCheck, setVarsCheck in *;
unfold bop_op in *;
simpl in *).

End C.

(******************************************************************)
(* All the infividual C machines are defined here *)
(* Perhaps moving this elsewhere would be a good idea *)

Require Import pm.
Require Import pe.
Require Import ald.
Require Import ale.
Require Import hw.
Require Import pmap.
Require Import "as".


Module C_PM := C (PM_State).
Module C_PE := C (PE_State).
Module C_ALD := C (ALD_State).
Module C_ALE := C (ALE_State).
Module C_PMAP := C (PMAP_State).
Module C_HW := C (HW_State).
Module C_AS := C (AS_State).
