(* mem_sig_ald.v *)
(* Specifications for the memory allocator over the PD memory model *)
(* Also includes the PD memory model specification for mem_init and a stub for pt_init *)

Require Import List.
Require Import Bool.
Require Import NArith.
Require Import ZArith.

Require Import spec.
Require Import label.

Require Import zbool_assist.
Require Import base_simp.
Require Import Omega.
Require Import Program.

Require Import c.

Require Import sact.
Require Import actsmp.

Require Import pe.

(* Needed for the definition of PTdom *)
Require Import ale.

Module mem_sig_ald.
Import C_PM SCAP.

Notation "a ::: b" := (Iseq a b) (at level 80).
Notation "a ;; b" := (ActionChain a b) (at level 80).

Definition getPageInfoCheck (pg : Z) : SCheck (State := Mem.t) := fun s => ValidPg pg.

Definition getPageInfo (pg : Z) : SGet bool :=
  fun s => if Zeq_bool (getMem (PE_State.PMM_data_loc + pg * 8) s) 0%Z then false else true.

Definition setPageInfoCheck (pg : Z) : SCheck (State := Mem.t) := fun s => ValidPg pg.

Definition setPageInfo (pg :Z) (pi : bool) : SAct (State := Mem.t) :=
  setMem (PE_State.PMM_data_loc + pg * 8) (if pi then 1%Z else 0%Z).
	
Definition PMM_OK (s : option State) : Prop := True.
Definition PMM_valid (s : Mem.t) : Prop := True.

Definition mem_free_spec : Action State :=
{| pre := (fun s => StackCheck (getArgsCheck 1) s = true /\ 
                  let pg := nth 0 (StackGet getArgs s) 0%Z in 
                  (0 < pg < NPAGES)%Z /\
                  MemGet (getPageInfo pg) s = true
	      	);

     rel := (fun s s' => let pg := nth 0 (StackGet getArgs s) 0%Z in
                 if MemGet (getPageInfo pg) s then
                         s' = (StackAct (pop +++ setRet 0%Z)  +++ (MemAct (setPageInfo pg false))) s
                 else False
               )
|}.

Definition mem_alloc_spec : Action State :=
{|
pre := (fun s => 
     PMM_OK (Some s) /\ PMM_valid (s.(memory)) /\ StackCheck (getArgsCheck 0) s = true 
                /\ StackCheck isFrame s = true /\ (1 < NPAGES)%Z
     );

rel := (fun s s' =>
   (exists page, (1 <= page < NPAGES)%Z /\  MemGet (getPageInfo page) s = false /\ 
          (s' = (StackAct (pop +++ setRet page) +++ MemAct (setPageInfo page true)) s))
   \/
   (s' = StackAct (pop +++ setRet 0%Z) s))
|}.

Definition mem_spec : SpecHeap := fun l =>
  match l with
  | mem_free_label => Some mem_free_spec
  | mem_alloc_label => Some mem_alloc_spec
  | _ => None
  end.

(***************************************************)
(* mem_init *)

Definition PM_mem_init_spec : Action C_PM.State.t :=
 {| pre := (fun (s : C_PM.State.t) => C_PM.State.StackCheck (getArgsCheck 0) s = true); 
   rel := fun (s s' : C_PM.State.t) => 
      (forall l, ALE_State.PtDom l = true -> 
                      s'.(C_PM.State.memory) (PE_State.PMM_data_loc + Pg l * 8)%Z = 1%Z) /\
     (forall l, PE_State.PMMDom l = true -> 
                      s'.(C_PM.State.memory) (PE_State.PMM_data_loc + Pg l * 8)%Z = 1%Z) /\
       (forall l : Z,
                ValidPA l = true ->
                (s'.(C_PM.State.memory) (PE_State.PMM_data_loc + Pg l * 8)%Z = 1%Z \/
                s'.(C_PM.State.memory) (PE_State.PMM_data_loc + Pg l * 8)%Z = 0%Z)
        ) /\
     s'.(C_PM.State.stack) = setRet 0 (pop s.(C_PM.State.stack))
|}. 



Definition lib_mem_init_spec : C_PM.SCAP.SpecHeap.SpecHeap := fun l =>
  match l with
  | mem_init_label => Some PM_mem_init_spec
  | _ => None
  end.

(*****************************************************)
(* Stubs for pt_init *)

Definition PM_pmap_init_spec : Action C_PM.State.t :=
 {| pre := (fun (s : C_PM.State.t) => C_PM.State.StackCheck (getArgsCheck 0) s = true /\ 
       (forall l : Z,
                ValidPA l = true ->
                (s.(C_PM.State.memory) (PE_State.PMM_data_loc + Pg l * 8)%Z = 1%Z \/
                s.(C_PM.State.memory) (PE_State.PMM_data_loc + Pg l * 8)%Z = 0%Z)
        ) /\
                                          (forall l, ValidPA l = true -> PE_State.PMMDom l = true -> 
                                                       s.(C_PM.State.memory) (PE_State.PMM_data_loc + Pg l * 8)%Z = 1%Z) /\
                                          (forall l, ValidPA l = true -> ALE_State.PtDom l = true -> 
                                                       s.(C_PM.State.memory) (PE_State.PMM_data_loc + Pg l * 8)%Z = 1%Z)
                                           ); 
   rel := fun (s s' : C_PM.State.t) => 
     (forall pg, ValidPg pg = true -> s'.(C_PM.State.memory) (PTROOT + pg * 8)%Z = pg) /\
     (forall pg, ValidVPg pg = true -> ValidPg pg = false -> 
                            s'.(C_PM.State.memory) (PTROOT + pg * 8)%Z= 0%Z) /\

          (forall pg, ValidPg pg = true -> 
                           s'.(C_PM.State.memory) (PE_State.PMM_data_loc + pg * 8)%Z = 
                           s.(C_PM.State.memory) (PE_State.PMM_data_loc + pg * 8)%Z
           ) /\
     s'.(C_PM.State.stack) = setRet 0 (pop s.(C_PM.State.stack))
|}. 


End mem_sig_ald.

