(* actsmp.v *)
(* Contains lemmas and tactics used to simplify verification *)
(* by doing targeted unfoldings, clean up, etc *)
(* Can benefit from a good clean up and overhaul *)

Require Import List.
Require Import Bool.
Require Import NArith.
Require Import ZArith.
Require Import IfProp.

Require Import spec.
Require Import scap.


Require Import zbool_assist.
Require Import Omega.

Require Import Logic.Classical_Prop.
Require Import Logic.FunctionalExtensionality.


Ltac detect b:=
match goal with
  | h1 : _ |- _ => match (constr:b) with | h1 => idtac "found" h1 end
end.

Ltac detect' b f :=
progress (
idtac "looking for" b;
first [
match goal with
  | h1 : _ |- _ => match (constr:b) with | h1 => idtac "match found" end
end | (idtac "making new" b; f b)]).


Ltac autosolve := Zeq_norm; try solve [auto with mcnc | easy | omega | elimtype False; omega | 
elimtype False; eauto with mcnc].

Lemma break_andb : forall a b, a && b = true -> (a = true) /\ (b = true).
intros.
destruct a; destruct b; firstorder.
Qed.

Lemma break_andb_goal : forall a b, (a = true) -> (b = true) -> (a && b = true).
intros; destruct a; destruct b; firstorder.
Qed.

Ltac autoclean :=
match goal with
 | h1 : True |- _ => clear h1
 | h1 : ?a = ?a |- _ => clear h1
 | h1 : _ /\ _ |- _ => destruct h1
 | h1 : ex _ |- _ => destruct h1
 | |- forall x, _ => intros
 | |- _ -> _ => intros
 | h1 : (?a _) = (?a _) |- _ => inversion h1; clear h1; try subst; try solve [autosolve]
 | h1 : ?a && ?b = true |- _ => specialize (break_andb _ _ h1); clear h1; intro h1
 | h1 : match ?a with | Some _ => _ | None => False end |- _ => 
           detect' a makenew'; try solve [autosolve]
 | h1 : _ = _ |- _ => progress subst
 | h1 : True -> _ |- _ => specialize (h1 I)
end.

Lemma pre_ActionCheck_hyp : forall State p g (s : State),
  pre (ActionCheck p g) s -> (p s /\ pre g s).
firstorder.
Qed.

Lemma pre_ActionChain_goal : forall State g g' (s : State),
pre g s -> (forall s' : State, rel g s s' -> pre g' s') ->
pre (ActionChain g g') s.
firstorder.
Qed.

Lemma pre_ActionPlus_hyp : forall State p g g' (s : State),
  pre (ActionPlus p g g') s -> (pre g s /\ p s \/ pre g' s /\ ~ p s).
firstorder.
Qed.

Lemma rel_ActionChain_hyp : forall State g g' (s s' : State),
rel (ActionChain g g') s s' -> (exists s'', rel g s s'' /\ rel g' s'' s').
firstorder.
Qed.

Lemma rel_ActionPlus_hyp : forall State p g g' (s s' : State),
rel (ActionPlus p g g') s s' -> 
((p s /\ rel g s s') \/ (~ p s /\ rel g' s s')).
firstorder.
destruct (classic (p s)); firstorder.
Qed.

Lemma rel_ActionChain_order : forall State g g' g'' (s s' : State),
rel (ActionChain (ActionChain g g') g'') s s' -> rel (ActionChain g (ActionChain g' g'')) s s'.
unfold ActionChain; simpl; intros.
destruct H; firstorder.
Qed.

Lemma pre_ActionPlus_goal : forall State p g g' (s : State),
  (pre g s /\ p s \/ pre g' s /\ ~ p s) ->
  pre (ActionPlus p g g') s.
firstorder.
Qed.

Lemma rel_actionPlus_goal : forall State (p : State -> Prop) g g' (s s' : State),
  (p s -> rel g s s') -> ((~ p s) -> rel g' s s') ->
  rel (ActionPlus p g g') s s'.
firstorder.
Qed.

Lemma pre_ActionLetO_hyp : forall State A get f s, 
  pre (ActionLetO (get: State -> option A)  f) (s : State) -> 
  (match get s with
    | Some x => pre (f x) s
    | None => False
    end).
firstorder.
Qed.

Lemma rel_ActionLetO_goal : forall State A get f (s s' : State),
  (match get s with
    | Some x => rel (f x) s s'
    | None => False
    end) ->
    rel (ActionLetO (get: State -> option A) f) s s'.
firstorder.
Qed.

Lemma match_state_goal : forall State s (s0 : State) sx,
Some s0 = sx ->
s = sx ->
match s 
with
| Some s'' => s0 = s''
| None => False
end.
firstorder.
subst; auto.
Qed.


Ltac structured :=
match goal with 
 | h1 : pre (ActionCheck _ _) _ |- _ => 
    idtac "ActionCheck hyp"; 
    specialize (pre_ActionCheck_hyp _ _ _ _ h1); clear h1; intros h1; destruct h1
 | h1 : pre (ActionPlus _ _ _) _ |- _ =>
    idtac "pre ActionPlus hyp"; 
    specialize (pre_ActionPlus_hyp _ _ _ _ _ h1); clear h1; intros h1; destruct h1
 | h1 : pre (ActionLetO ?get _) ?s |- _ =>
    idtac "pre ActionLetO hyp";
    specialize (pre_ActionLetO_hyp _ _ _ _ _ h1); clear h1; intros h1; idtac get s;
    makenew' (get s); autosolve
 | h1 : rel (ActionPlus _ _ _) _ _ |- _ =>
    idtac "rel ActionPlus hyp"; 
    specialize (rel_ActionPlus_hyp _ _ _ _ _ _ h1); clear h1; intros h1; destruct h1     
 | h1 : rel (ActionChain (ActionChain _ _) _) _ _ |- _ => 
    idtac "rel ActionChain2 hyp normalize"; 
    specialize (rel_ActionChain_order _ _ _ _ _ _ h1); clear h1; intros h1
   
 | h1 : rel (ActionChain _ _) _ _ |- _ =>
    idtac "rel ActionChain hyp"; 
    specialize (rel_ActionChain_hyp _ _ _ _ _ h1); clear h1; intros h1; 
    let x := fresh in (destruct h1 as [x h1]; destruct h1) 
 | |- pre (ActionChain _ _) _ => idtac "ActionChain goal"; apply pre_ActionChain_goal
 | |- pre (ActionPlus _ _ _) _ => idtac "pre ActionPlus goal"; apply pre_ActionPlus_goal
 | |- rel (ActionPlus _ _ _) _ _ => idtac "rel ActionPlus goal"; apply rel_actionPlus_goal
 | |- rel (ActionLetO ?get _) ?s ?s' =>
    idtac "rel ActionLetO goal"; apply rel_ActionLetO_goal
 | h1 : Some ?s0 = _ |- (match _ with | Some sx => ?s0 = sx | None => False end) => eapply match_state_goal; eauto
end.


