(* refinements.v *)
(* Defines refinements that are can be defined as functions from concrete machine to abstract machine *)


Require Import Equalities.
Require Import scap.
Require Import spec.
Require Import zbool_assist.


Module Refinement (AMachineState : Typ) (AMachineOp : Typ) (ADecider : Typ)
(CMachineState : Typ) (CMachineOp : Typ) (CDecider : Typ).

Module ASCAP := SCAP AMachineState AMachineOp ADecider.
Module CSCAP := SCAP CMachineState CMachineOp CDecider.

Definition AState := AMachineState.t.
Definition CState := CMachineState.t.

Import ASCAP.
Import ASCAP.Process.
Import CSCAP.
Import CSCAP.Process.

Definition Refinement 
   (SpecRel : ASCAP.SpecHeap.SpecHeap -> CSCAP.SpecHeap.SpecHeap -> Prop)
   (CodeRel : ASCAP.CodeHeap -> CSCAP.CodeHeap -> Prop)
   Adp Cdp AMach CMach :=
   forall AC APsi AHypo CC CPsi CHypo,
   SpecRel AHypo CHypo ->
   SpecRel APsi CPsi -> 
   CodeRel AC CC ->
   ASCAP.WFCode Adp AMach AHypo AC APsi -> 
   CSCAP.WFCode Cdp CMach CHypo CC CPsi.


Definition specrel 
     (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label)
     (ActionRel : Action AState -> Action CState -> Prop) : 
                   ASCAP.SpecHeap.SpecHeap -> CSCAP.SpecHeap.SpecHeap -> Prop :=
    fun APsi CPsi =>
   (forall l, match APsi l with
      | Some Ag => match CPsi (LabelConv l) with
                   | Some Cg => ActionRel Ag Cg
                   | None => False
                   end
      | None => match CPsi (LabelConv l) with
                   | Some _ => False
                   | None => True
                   end
      end
   ).

Definition coderel
     (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label)
     (ProcRel : ASCAP.Process.Process -> CSCAP.Process.Process -> Prop) : 
                   ASCAP.CodeHeap -> CSCAP.CodeHeap -> Prop :=
    fun AC CC =>
   (forall l, match AC l with
      | Some AI => match CC (LabelConv l) with
                   | Some CI => ProcRel AI CI
                   | None => False
                   end
      | None => True
      end
   ).
   
Definition ProcRefinement
   (ActionRel : Action AState -> Action CState -> Prop)
   (ProcRel : ASCAP.Process.Process -> CSCAP.Process.Process -> Prop)
   (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label)
   Adp Cdp AMach CMach :=
   forall APsi Ag AI CPsi Cg CI,
   specrel LabelConv ActionRel APsi CPsi ->
   ActionRel Ag Cg ->
   ProcRel AI CI ->
   ASCAP.Process.WFProc Adp AMach APsi Ag AI -> 
   CSCAP.Process.WFProc Cdp CMach CPsi Cg CI.


Ltac make a :=
        let newdecl := fresh in
        let neweq := fresh in
        saveas a newdecl neweq.

Lemma ProcRefinementValid 
  (LabelConv' : CSCAP.Process.Label -> ASCAP.Process.Label)
  (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label) 
  (LL : forall l, LabelConv (LabelConv' l) = l) :
  forall ActionRel ProcRel Adp Cdp AMach CMach,
   ProcRefinement ActionRel ProcRel LabelConv Adp Cdp AMach CMach ->
   Refinement (specrel LabelConv ActionRel)
              (coderel LabelConv ProcRel)
              Adp Cdp AMach CMach.
Proof.
intros.
unfold Refinement.
unfold ProcRefinement in *.
intros.

red in H0; red in H1; red in H2.
inversion_clear H3.

assert (SpecHeap_disjoint CHypo CPsi).
clear H4.
red in disj.
red; intros.
specialize (disj (LabelConv' l)).
specialize (H0 (LabelConv' l)).
specialize (H1 (LabelConv' l)).
rewrite LL in H0, H1.
makenew' (CHypo l); auto.
makenew' (AHypo (LabelConv' l)); try solve [inversion H0].
makenew' (APsi (LabelConv' l)); try solve [inversion H1].
inversion disj.
makenew' (CPsi l); try contradiction; auto.
rename H3 into newdisj.

apply WFCode_1 with newdisj; intros.
set (SR := H1); generalize SR; clear SR; intros SR.
(* specialize H0 with (LabelConv' f). *)
specialize H1 with (LabelConv' f).
specialize H2 with (LabelConv' f).
specialize (H4 (LabelConv' f)).
unfold ASCAP.Process.getLabelAction in H4.
red in H3.
rewrite LL in H1.
rewrite H3 in H1.
makenew' (APsi (LabelConv' f)).
specialize H4 with a; firstorder.
rewrite H4 in H2.
rewrite LL in H2.
makenew' (CC f); firstorder.
exists p; firstorder.

apply H with (ASCAP.SpecHeap.SpecHeap_join AHypo APsi disj) a x; auto.
red; intros.
unfold ASCAP.SpecHeap.SpecHeap_join; simpl.
specialize (H0 l).
makenew' (AHypo l); simpl; auto.
unfold SpecHeap_join.
makenew' (CHypo (LabelConv l)); simpl; try contradiction; auto.
specialize (SR l).
makenew' (APsi l).
unfold SpecHeap_join.
makenew' (CHypo (LabelConv l)); simpl; auto.
makenew' (CPsi (LabelConv l)); simpl; try contradiction.
unfold SpecHeap_join.
makenew' (CHypo (LabelConv l)); simpl; try contradiction.
makenew' (CPsi (LabelConv l)); simpl; try contradiction.
auto.

rewrite LL in H2.
specialize SR with (LabelConv' f).
rewrite H6 in SR.
rewrite LL in SR.
rewrite H3 in SR.
contradiction.
Qed.



Definition OrderRefinement
   (InstrRel : ASCAP.Process.Op -> CSCAP.Process.Op)
   (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
   (ActionRel : Action AState -> Action CState -> Prop)
   (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label)
   Adp Cdp
   (AMach : ASCAP.Process.Machine) (CMach : CSCAP.Process.Machine) :=
   forall APsi CPsi,
   specrel LabelConv ActionRel APsi CPsi ->
   (forall Ag, exists Cg, ActionRel Ag Cg) /\
   (forall Ai, match AMach Ai with 
               | Some _ => exists Cig, CMach (InstrRel Ai) = Some Cig 
               | None => True 
               end) /\
   (forall Ai Ci Aig Cig Ag Cg,
       InstrRel Ai = Ci -> AMach Ai = Some Aig -> CMach Ci = Some Cig ->
       ActionRel Ag Cg -> ActionGTE Ag Aig -> ActionGTE Cg Cig) /\
   (forall Ag Cg Ag' Cg',
       ActionRel Ag Cg -> ActionRel Ag' Cg' -> ActionGTE Ag Ag' -> ActionGTE Cg Cg') /\
   (forall Ag Cg, 
       ActionRel Ag Cg -> ActionGTE Ag (ActionID AState) -> ActionGTE Cg (ActionID CState)) /\
   (forall d Ag Ag' Cg Cg' Cg'',
       ActionRel Ag Cg -> ActionRel Ag' Cg' -> ActionRel (ActionPlus (Adp d) Ag Ag') Cg'' ->
       ActionGTE Cg'' (ActionPlus (Cdp (DeciderRel d)) Cg Cg')) /\
   (forall Ag Ag' Cg Cg' Cg'',
       ActionRel Ag Cg -> ActionRel Ag' Cg' -> ActionRel (ActionChain Ag Ag') Cg'' ->
       ActionGTE Cg'' (ActionChain Cg Cg')).


Section procrel.
Variable oprel : ASCAP.Process.Op -> CSCAP.Process.Op.
Variable labelconv : ASCAP.Process.Label -> CSCAP.Process.Label.
Variable DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider.
Inductive procrel
    : ASCAP.Process.Process -> CSCAP.Process.Process -> Prop :=
  | procrel_nil : procrel ASCAP.Process.Inil CSCAP.Process.Inil
  | procrel_perf : forall i,
       procrel (ASCAP.Process.Iperf i) (CSCAP.Process.Iperf (oprel i))
  | procrel_call : forall l,
       procrel (ASCAP.Process.Icall l) (CSCAP.Process.Icall (labelconv l))
  | procrel_plus : forall p AI1 AI2 CI1 CI2,
       procrel AI1 CI1 -> procrel AI2 CI2 ->
       procrel (ASCAP.Process.Iplus p AI1 AI2) (CSCAP.Process.Iplus (DeciderRel p) CI1 CI2)
  | procrel_seq : forall AI1 AI2 CI1 CI2,
       procrel AI1 CI1 -> procrel AI2 CI2 ->
       procrel (ASCAP.Process.Iseq AI1 AI2) (CSCAP.Process.Iseq CI1 CI2)
.
End procrel.

Lemma OrderRefinementValid 
  (LabelConv' : CSCAP.Process.Label -> ASCAP.Process.Label)
  (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label) 
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
  (LL : forall l, LabelConv (LabelConv' l) = l) :
  forall ActionRel InstrRel Adp Cdp AMach CMach,
   OrderRefinement InstrRel DeciderRel ActionRel LabelConv Adp Cdp AMach CMach ->
   ProcRefinement ActionRel (procrel InstrRel LabelConv DeciderRel) LabelConv Adp Cdp AMach CMach.

Proof.
intros.
red; intros.
generalize CI Cg H2 H1; clear H1 H2 Cg CI.
induction H3;
red in H.

(* nil *)
specialize H with Psi CPsi.
firstorder.
inversion_clear H2.
apply WFProc_weaken with (ActionID CState); auto.
apply H6 with (ActionID AState); auto.
apply ActionGTE_refl.

(* perf *)
specialize H with Psi CPsi.
firstorder.
inversion_clear H2.
red in H1.
make (CMach (InstrRel i)).
destruct H2 as [Cig | ].
apply WFProc_weaken with Cig.
apply H5 with i (InstrRel i) g g; auto.
apply ActionGTE_refl.
constructor; auto.
specialize H4 with i.
rewrite H1 in H4.
rewrite H10 in H4.
inversion H4.
inversion H2.

(* call *)
specialize H with Psi CPsi.
firstorder.
inversion_clear H2.
red in H1.
red in H0.
specialize H0 with l.
rewrite H1 in H0.
makenew' (CPsi (LabelConv l)); try contradiction.
apply WFProc_weaken with a; auto.
apply H6 with g g; auto.
apply ActionGTE_refl.

(* weaken *)
set (y := H); generalize y; clear y; intros.
specialize H with Psi CPsi.
destruct H; [firstorder | ].
specialize H with g'.
inversion H.
apply WFProc_weaken with x; auto.
decompose record H5; clear H5.
apply H8 with g g'; auto.

(* chain *)
set (y := H); generalize y; clear y; intros.
specialize H with Psi CPsi.
destruct H; [firstorder | ].
decompose record H3; clear H3.
inversion_clear H2.
set (xxx := H g').
destruct xxx as [Cg'].
set (xxx := H g'').
destruct xxx as [Cg''].
apply WFProc_weaken with (ActionChain Cg' Cg''); auto.
apply H10 with g' g''; auto.

(* plus *)
set (y := H); generalize y; clear y; intros.
specialize H with Psi CPsi.
destruct H; [firstorder | ].
decompose record H3; clear H3.
inversion_clear H2.
set (xxx := H g').
destruct xxx as [Cg'].
set (xxx := H g'').
destruct xxx as [Cg''].
apply WFProc_weaken with (ActionPlus (Cdp (DeciderRel p)) Cg' Cg''); auto.
apply H8 with g' g''; auto.
Qed.


Lemma OrderRefinementValid2
  (LabelConv' : CSCAP.Process.Label -> ASCAP.Process.Label)
  (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label) 
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
  (LL : forall l, LabelConv (LabelConv' l) = l) :
  forall ActionRel InstrRel Adp Cdp AMach CMach,
   OrderRefinement InstrRel DeciderRel ActionRel LabelConv Adp Cdp AMach CMach ->
    Refinement (specrel LabelConv ActionRel)
              (coderel LabelConv (procrel InstrRel LabelConv DeciderRel))
              Adp Cdp AMach CMach.
Proof.
intros.
eapply ProcRefinementValid; eauto.
eapply OrderRefinementValid; eauto.
Qed.













Definition OrderRefinement2
   (InstrRel : ASCAP.Process.Op -> CSCAP.Process.Op)
   (ActionRel : Action AState -> Action CState -> Prop)
   (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label)
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
   Adp Cdp (AMach : ASCAP.Process.Machine) (CMach : CSCAP.Process.Machine) :=
   forall APsi CPsi,
   specrel LabelConv ActionRel APsi CPsi ->
   (forall Ag, exists Cg, ActionRel Ag Cg) /\
   (forall Ai, match AMach Ai with 
               | Some Aig => exists Cig, CMach (InstrRel Ai) = Some Cig /\ 
                          (forall x, ActionRel Aig x -> ActionGTE x Cig)
               | None => True 
               end) /\
   (forall Ag Cg Ag' Cg',
       ActionRel Ag Cg -> ActionRel Ag' Cg' -> ActionGTE Ag Ag' -> ActionGTE Cg Cg') /\
   (forall Ag Cg, 
       ActionRel Ag Cg -> ActionGTE Ag (ActionID AState) -> ActionGTE Cg (ActionID CState)) /\
   (forall p Ag Ag' Cg Cg' Cg'',
       ActionRel Ag Cg -> ActionRel Ag' Cg' -> ActionRel (ActionPlus (Adp p) Ag Ag') Cg'' ->
       ActionGTE Cg'' (ActionPlus (Cdp (DeciderRel p)) Cg Cg')) /\
   (forall Ag Ag' Cg Cg' Cg'',
       ActionRel Ag Cg -> ActionRel Ag' Cg' -> ActionRel (ActionChain Ag Ag') Cg'' ->
       ActionGTE Cg'' (ActionChain Cg Cg')).


Lemma OrderRefinementValid3
  (LabelConv' : CSCAP.Process.Label -> ASCAP.Process.Label)
  (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label) 
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
  (LL : forall l, LabelConv (LabelConv' l) = l) :
  forall ActionRel InstrRel Adp Cdp AMach CMach,
   OrderRefinement2 InstrRel ActionRel LabelConv DeciderRel Adp Cdp AMach CMach ->
   ProcRefinement ActionRel (procrel InstrRel LabelConv DeciderRel) LabelConv Adp Cdp AMach CMach.
intros.
red; intros.
generalize CI Cg H2 H1; clear H1 H2 Cg CI.
induction H3;
red in H.

(* nil *)
specialize H with Psi CPsi.
firstorder.
inversion_clear H2.
apply WFProc_weaken with (ActionID CState); auto.
apply H5 with (ActionID AState); auto.
apply ActionGTE_refl.

(* perf *)
specialize H with Psi CPsi.
firstorder.
inversion_clear H2.
red in H1.
specialize H4 with i.
rewrite H1 in H4.
destruct H4; intuition.
apply WFProc_weaken with x; auto.

(* call *)
specialize H with Psi CPsi.
firstorder.
inversion_clear H2.
red in H1.
red in H0.
specialize H0 with l.
rewrite H1 in H0.
makenew' (CPsi (LabelConv l)); try contradiction.
apply WFProc_weaken with a; auto.
apply H5 with g g; auto.
apply ActionGTE_refl.

(* weaken *)
set (y := H); generalize y; clear y; intros.
specialize H with Psi CPsi.
destruct H; [firstorder | ].
specialize H with g'.
inversion H.
apply WFProc_weaken with x; auto.
decompose record H5; clear H5.
apply H9 with g g'; auto.

(* chain *)
set (y := H); generalize y; clear y; intros.
specialize H with Psi CPsi.
destruct H; [firstorder | ].
decompose record H3; clear H3.
inversion_clear H2.
set (xxx := H g').
destruct xxx as [Cg'].
set (xxx := H g'').
destruct xxx as [Cg''].
apply WFProc_weaken with (ActionChain Cg' Cg''); auto.
apply H9 with g' g''; auto.

(* plus *)
set (y := H); generalize y; clear y; intros.
specialize H with Psi CPsi.
destruct H; [firstorder | ].
decompose record H3; clear H3.
inversion_clear H2.
set (xxx := H g').
destruct xxx as [Cg'].
set (xxx := H g'').
destruct xxx as [Cg''].
apply WFProc_weaken with (ActionPlus (Cdp (DeciderRel p)) Cg' Cg''); auto.
apply H7 with g' g''; auto.
Qed.











(************************************************************************)

Definition liftAction (f : CState -> AState) (* (f' : AState -> CState) *): 
     Action AState -> Action CState := fun g =>
mkAct (fun s => g.(pre) (f s))
      (fun s s' => g.(rel) (f s) (f s')).

Lemma ActionGTE_lift : forall f Ag Ag', 
    ActionGTE Ag Ag' -> ActionGTE (liftAction f Ag) (liftAction f Ag').
firstorder.
Qed.

Lemma ActionGTE_lift_ID : forall f', 
    ActionGTE (liftAction f' (ActionID AState)) (ActionID CState).
    firstorder; simpl in *.
    rewrite H0; auto.
Qed.

Lemma ActionGTE_lift_Plus : forall f' p 
       (adp : ASCAP.Process.Decider_to_Prop) (cdp : CSCAP.Process.Decider_to_Prop) Ag Ag' Cg Cg'
      (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
      (DeciderLift : forall d s, (adp d (f' s)) -> (cdp (DeciderRel d) s))
      (DeciderLift' : forall d s, (cdp (DeciderRel d) s) -> (adp d (f' s)))
,
      ActionGTE (liftAction f' Ag) Cg ->
      ActionGTE (liftAction f' Ag') Cg' ->
      ActionGTE (liftAction f' (ActionPlus (adp p) Ag Ag')) 
                           (ActionPlus (cdp (DeciderRel p)) Cg Cg').
unfold liftAction, ActionGTE; simpl; intuition.
Qed.

Lemma ActionGTE_lift_Chain : forall f' Ag Ag' Cg Cg',
      ActionGTE (liftAction f' Ag) Cg ->
      ActionGTE (liftAction f' Ag') Cg' ->
      ActionGTE (liftAction f' (ActionChain Ag Ag')) (ActionChain Cg Cg').
firstorder.
unfold ActionChain; simpl.
exists (f' x); firstorder.
specialize (H2) with x s'; firstorder.
Qed.







Definition actionrel (f : CState -> AState) 
            : Action AState -> Action CState -> Prop :=
  fun (Ag : Action AState) (Cg : Action CState) => 
       ActionGTE Cg (liftAction f Ag) /\  ActionGTE (liftAction f Ag) Cg.


Definition IsoRefinement
   (InstrRel : ASCAP.Process.Op -> CSCAP.Process.Op)
   (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label)
   (AMach : ASCAP.Process.Machine) (CMach : CSCAP.Process.Machine) 
   (f : CState -> AState) 
   :=
   forall APsi CPsi,
   specrel LabelConv (actionrel f) APsi CPsi ->
   (forall Ai Aig, 
          AMach Ai = Some Aig -> 
          exists Cig, CMach (InstrRel Ai) = Some Cig /\
             ActionGTE (liftAction f Aig) Cig).

Lemma IsoRefinementValid
  (LabelConv' : CSCAP.Process.Label -> ASCAP.Process.Label)
  (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label)
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
  (Adp : ASCAP.Process.Decider_to_Prop)
  (Cdp : CSCAP.Process.Decider_to_Prop)
  (LL : forall l, LabelConv (LabelConv' l) = l)
  (f : CState -> AState)
  (DeciderLift : forall d s, (Adp d (f s)) -> (Cdp (DeciderRel d) s))
  (DeciderLift' : forall d s, (Cdp (DeciderRel d) s) -> (Adp d (f s)))
:
  forall InstrRel AMach CMach,
   IsoRefinement InstrRel LabelConv AMach CMach f -> 
   OrderRefinement2 InstrRel (actionrel f) LabelConv DeciderRel Adp Cdp AMach CMach.

Proof.
  intros.
  red; intros.
  red in H.
  specialize (H APsi CPsi); intuition.
  exists (liftAction f Ag); firstorder.

  makenew' (AMach Ai); firstorder.

  unfold actionrel in *.
  intuition.
  apply ActionGTE_trans with (liftAction f Ag'); auto.
  apply ActionGTE_trans with (liftAction f Ag); auto.
  apply ActionGTE_lift; auto.

  unfold actionrel in *; intuition.
  apply ActionGTE_trans with (liftAction f Ag); auto.
  apply ActionGTE_trans with (liftAction f (ActionID AState)); auto.
  apply ActionGTE_lift; auto.
  apply ActionGTE_lift_ID; auto.

  unfold actionrel in *; intuition.
  apply ActionGTE_trans with (liftAction f (ActionPlus (Adp p) Ag Ag')); auto.
  apply ActionGTE_lift_Plus; auto.

  unfold actionrel in *; intuition.
  apply ActionGTE_trans with (liftAction f (ActionChain Ag Ag')); auto.
  apply ActionGTE_lift_Chain; auto.
Qed.


Lemma IsoRefinementValid'
  (LabelConv' : CSCAP.Process.Label -> ASCAP.Process.Label)
  (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label) 
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
  (Adp : ASCAP.Process.Decider_to_Prop)
  (Cdp : CSCAP.Process.Decider_to_Prop)
  (LL : forall l, LabelConv (LabelConv' l) = l)
  (f : CState -> AState)
  (DeciderLift : forall d s, (Adp d (f s)) -> (Cdp (DeciderRel d) s))
  (DeciderLift' : forall d s, (Cdp (DeciderRel d) s) -> (Adp d (f s)))
:
  forall InstrRel AMach CMach,
   IsoRefinement InstrRel LabelConv AMach CMach f -> 
   OrderRefinement InstrRel DeciderRel (actionrel f) LabelConv Adp Cdp AMach CMach.

Proof.
  intros.
  red; intros.
  red in H.
  specialize (H APsi CPsi H0); intuition.
  exists (liftAction f Ag); firstorder.

  makenew' (AMach Ai); firstorder.

  specialize (H Ai Aig H2).
  unfold actionrel in *.
  destruct H.
  inversion H.
  rewrite H1 in H6.
  rewrite H3 in H6.
  inversion H6.
  subst.
  intuition.
  apply ActionGTE_trans with (liftAction f Aig); auto.
  apply ActionGTE_trans with (liftAction f Ag); auto.
  apply ActionGTE_lift; auto.

  unfold actionrel in *.
  intuition.
  apply ActionGTE_trans with (liftAction f Ag'); auto.
  apply ActionGTE_trans with (liftAction f Ag); auto.
  apply ActionGTE_lift; auto.

  unfold actionrel in H1; intuition.
  apply ActionGTE_trans with (liftAction f Ag); auto.
  apply ActionGTE_trans with (liftAction f (ActionID AState)); auto.
  apply ActionGTE_lift; auto.
  apply ActionGTE_lift_ID; auto.

  unfold actionrel in *; intuition.
  apply ActionGTE_trans with (liftAction f (ActionPlus (Adp d) Ag Ag')); auto.
  apply ActionGTE_lift_Plus; auto.

  unfold actionrel in *; intuition.
  apply ActionGTE_trans with (liftAction f (ActionChain Ag Ag')); auto.
  apply ActionGTE_lift_Chain; auto.
Qed.





(**********************************)

Definition actionrel' (f : CState -> AState) 
            : Action AState -> Action CState -> Prop :=
  fun (Ag : Action AState) (Cg : Action CState) => (liftAction f Ag) = Cg.


Definition IsoRefinement'
   (InstrRel : ASCAP.Process.Op -> CSCAP.Process.Op)
   (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label)
   (AMach : ASCAP.Process.Machine) (CMach : CSCAP.Process.Machine) 
   (f : CState -> AState) 
   :=
   forall APsi CPsi,
   specrel LabelConv (actionrel' f) APsi CPsi ->
   (forall Ai Aig, 
          AMach Ai = Some Aig -> 
          exists Cig, CMach (InstrRel Ai) = Some Cig /\
           actionrel' f Aig Cig).

Lemma IsoRefinementValid''
  (LabelConv' : CSCAP.Process.Label -> ASCAP.Process.Label)
  (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label) 
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
  (Adp : ASCAP.Process.Decider_to_Prop)
  (Cdp : CSCAP.Process.Decider_to_Prop)
  (LL : forall l, LabelConv (LabelConv' l) = l)
  (f : CState -> AState) 
  (DeciderLift : forall d s, (Adp d (f s)) -> (Cdp (DeciderRel d) s))
  (DeciderLift' : forall d s, (Cdp (DeciderRel d) s) -> (Adp d (f s)))
:
  forall InstrRel AMach CMach,
   IsoRefinement' InstrRel LabelConv AMach CMach f -> 
   OrderRefinement InstrRel DeciderRel (actionrel' f) LabelConv Adp Cdp AMach CMach.

Proof.
  intros.
  red; intros.
  red in H.
  specialize (H APsi CPsi); intuition.
  exists (liftAction f Ag); firstorder.

  makenew' (AMach Ai); firstorder.

  red in H4; subst.
  specialize (H1 Ai Aig H2).
  inversion H1; intuition.
  red in H6; subst.
  rewrite H3 in H4; inversion H4; subst.
  apply ActionGTE_lift; auto.

  unfold actionrel' in *; subst.
  apply ActionGTE_lift; auto.

  unfold actionrel' in *; subst.
  apply ActionGTE_trans with (liftAction f (ActionID AState)); auto.
  apply ActionGTE_lift; auto.
  apply ActionGTE_lift_ID; auto.

  unfold actionrel' in *; subst.
  apply ActionGTE_lift_Plus; auto.
  apply ActionGTE_refl.
  apply ActionGTE_refl.

  unfold actionrel' in *; subst.
  apply ActionGTE_lift_Chain; auto.
  apply ActionGTE_refl.
  apply ActionGTE_refl.
Qed.

Lemma IsoRefinementValid'''
  (LabelConv' : CSCAP.Process.Label -> ASCAP.Process.Label)
  (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label) 
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
  (Adp : ASCAP.Process.Decider_to_Prop)
  (Cdp : CSCAP.Process.Decider_to_Prop)
  (LL : forall l, LabelConv (LabelConv' l) = l)
  (f : CState -> AState) 
  (DeciderLift : forall d s, (Adp d (f s)) -> (Cdp (DeciderRel d) s))
  (DeciderLift' : forall d s, (Cdp (DeciderRel d) s) -> (Adp d (f s)))
:
  forall InstrRel AMach CMach,
   IsoRefinement' InstrRel LabelConv AMach CMach f -> 
   OrderRefinement2 InstrRel (actionrel' f) LabelConv DeciderRel Adp Cdp AMach CMach.

Proof.
  intros.
  red; intros.
  red in H.
  specialize (H APsi CPsi H0); intuition.
  exists (liftAction f Ag); firstorder.

  makenew' (AMach Ai); firstorder.
  specialize (H Ai a H2); firstorder.
  exists x; split; auto.
  intros.
  unfold actionrel' in *; subst.
  apply ActionGTE_lift; auto.
  apply ActionGTE_refl.

  unfold actionrel' in *; subst.
  apply ActionGTE_lift; auto.

  unfold actionrel' in *; subst.
  apply ActionGTE_trans with (liftAction f (ActionID AState)); auto.
  apply ActionGTE_lift; auto.
  apply ActionGTE_lift_ID; auto.

  unfold actionrel' in *; subst.
  apply ActionGTE_lift_Plus; auto.
  apply ActionGTE_refl.
  apply ActionGTE_refl.

  unfold actionrel' in *; subst.
  apply ActionGTE_lift_Chain; auto.
  apply ActionGTE_refl.
  apply ActionGTE_refl.
Qed.



Lemma IsoRefinementValidTotal
  (LabelConv' : CSCAP.Process.Label -> ASCAP.Process.Label)
  (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label) 
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
  (Adp : ASCAP.Process.Decider_to_Prop)
  (Cdp : CSCAP.Process.Decider_to_Prop)
  (LL : forall l, LabelConv (LabelConv' l) = l)
  (f : CState -> AState)
  (DeciderLift : forall d s, (Adp d (f s)) -> (Cdp (DeciderRel d) s))
  (DeciderLift' : forall d s, (Cdp (DeciderRel d) s) -> (Adp d (f s)))
:
  forall InstrRel AMach CMach,
   IsoRefinement InstrRel LabelConv AMach CMach f -> 
   Refinement (specrel LabelConv (actionrel f))
              (coderel LabelConv (procrel InstrRel LabelConv DeciderRel))
              Adp Cdp AMach CMach.
intros.
eapply OrderRefinementValid2; eauto.
eapply IsoRefinementValid'; eauto.
Qed.

End Refinement.
