(* reprfinemenet.v *)
(* Defines representation based refinements *)

Require Import Equalities.
Require Import scap.
Require Import spec.
Require Import zbool_assist.

Set Implicit Arguments.

Module Refinement 
(AMachineState : Typ) (AMachineOp : Typ) (ADecider : Typ)
(ASCAP : SCAPt AMachineState AMachineOp ADecider)
(CMachineState : Typ) (CMachineOp : Typ) (CDecider : Typ)
(CSCAP : SCAPt CMachineState CMachineOp CDecider).

Definition AState := AMachineState.t.
Definition CState := CMachineState.t.

Import ASCAP.
Import CSCAP.

Section ReprFinement.


(* The entire system of REPR refinement is parametrized over a particular REPR *)
Variable repr : AState -> CState -> Prop.

(* This defines states that are related by REPR (adjoint states) *)
Definition Rrel (s s' : AState) : Prop := 
   exists cs, repr s cs /\ repr s' cs.


(* REPR related states must not be able to take separate branches  *)
(* The excl section defines this rule *)
Section excl.
Variable Adp : ASCAP.Process.Decider_to_Prop.
Definition excl (p : ASCAP.Process.Decider) := forall sa sa', Rrel sa sa' -> 
(Adp p sa -> ~~ Adp p sa').

Inductive excl_Proc : ASCAP.Process.Process -> Prop :=
  | excl_nil : excl_Proc ASCAP.Process.Inil
  | excl_perf : forall i, excl_Proc (ASCAP.Process.Iperf i)
  | excl_call : forall l, excl_Proc (ASCAP.Process.Icall l)
  | excl_seq : forall I1 I2, excl_Proc I1 -> excl_Proc I2 -> excl_Proc (ASCAP.Process.Iseq I1 I2)
  | excl_plus : forall p I1 I2, excl_Proc I1 -> excl_Proc I2 -> excl p ->
                   excl_Proc (ASCAP.Process.Iplus p I1 I2)
.

Definition excl_Code (C: ASCAP.CodeHeap) : Prop := forall f I, (C f = Some I) -> excl_Proc I.
End excl.

(********************************************************)

(* General Definition of refinement *)

Definition Refinement 
   (SpecRel : ASCAP.SpecHeap.SpecHeap -> CSCAP.SpecHeap.SpecHeap -> Prop)
   (CodeRel : ASCAP.CodeHeap -> CSCAP.CodeHeap -> Prop)
   Adp Cdp AMach CMach :=
   forall AC APsi AHypo CC CPsi CHypo,
   SpecRel AHypo CHypo ->
   SpecRel APsi CPsi -> 
   CodeRel AC CC ->
   ASCAP.WFCode Adp AMach AHypo AC APsi -> 
   CSCAP.WFCode Cdp CMach CHypo CC CPsi.

(********************************************************)

(* Definition of refinement that does not change procedures *)


Definition specrel 
     (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label)
     (ActionRel : Action AState -> Action CState -> Prop) : 
                   ASCAP.SpecHeap.SpecHeap -> CSCAP.SpecHeap.SpecHeap -> Prop :=
    fun APsi CPsi =>
   (forall l, match APsi l with
      | Some Ag => match CPsi (LabelConv l) with
                   | Some Cg => ActionRel Ag Cg
                   | None => False
                   end
      | None => match CPsi (LabelConv l) with
                   | Some _ => False
                   | None => True
                   end
      end
   ).

Definition coderel
     (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label)
     (ProcRel : ASCAP.Process.Process -> CSCAP.Process.Process -> Prop) : 
                   ASCAP.CodeHeap -> CSCAP.CodeHeap -> Prop :=
    fun AC CC =>
   (forall l, match AC l with
      | Some AI => match CC (LabelConv l) with
                   | Some CI => ProcRel AI CI
                   | None => False
                   end
      | None => True
      end
   ).
 

Definition ProcRefinement
   (ActionRel : Action AState -> Action CState -> Prop)
   (ProcRel : ASCAP.Process.Process -> CSCAP.Process.Process -> Prop)
   (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label)
   Adp Cdp AMach CMach :=
   forall APsi Ag AI CPsi Cg CI,
   specrel LabelConv ActionRel APsi CPsi ->
   ActionRel Ag Cg ->
   ProcRel AI CI ->
   ASCAP.Process.WFProc Adp AMach APsi Ag AI -> 
   CSCAP.Process.WFProc Cdp CMach CPsi Cg CI.


Ltac make a :=
        let newdecl := fresh in
        let neweq := fresh in
        saveas a newdecl neweq.

Lemma ProcRefinementValid 
  (LabelConv' : CSCAP.Process.Label -> ASCAP.Process.Label)
  (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label) 
  (LL : forall l, LabelConv (LabelConv' l) = l) :
  forall ActionRel ProcRel Adp Cdp AMach CMach,
   ProcRefinement ActionRel ProcRel LabelConv Adp Cdp AMach CMach ->
   Refinement (specrel LabelConv ActionRel)
              (coderel LabelConv ProcRel)
              Adp Cdp AMach CMach.
Proof.
intros.
unfold Refinement.
unfold ProcRefinement in *.
intros.

red in H0; red in H1; red in H2.
inversion_clear H3.

assert (SpecHeap_disjoint CHypo CPsi).
clear H4.
red in disj.
red; intros.
specialize (disj (LabelConv' l)).
specialize (H0 (LabelConv' l)).
specialize (H1 (LabelConv' l)).
rewrite LL in H0, H1.
makenew' (CHypo l); auto.
makenew' (AHypo (LabelConv' l)); try solve [inversion H0].
makenew' (APsi (LabelConv' l)); try solve [inversion H1].
inversion disj.
makenew' (CPsi l); try contradiction; auto.
rename H3 into newdisj.

apply WFCode_1 with newdisj; intros.
set (SR := H1); generalize SR; clear SR; intros SR.
specialize H1 with (LabelConv' f).
specialize H2 with (LabelConv' f).
specialize (H4 (LabelConv' f)).
unfold ASCAP.Process.getLabelAction in H4.
red in H3.
rewrite LL in H1.
rewrite H3 in H1.
makenew' (APsi (LabelConv' f)).
specialize H4 with a; firstorder.
rewrite H4 in H2.
rewrite LL in H2.
makenew' (CC f); firstorder.
exists p; firstorder.

apply H with (ASCAP.SpecHeap.SpecHeap_join AHypo APsi disj) a x; auto.
red; intros.
unfold ASCAP.SpecHeap.SpecHeap_join; simpl.
specialize (H0 l).
makenew' (AHypo l); simpl; auto.
unfold SpecHeap_join.
makenew' (CHypo (LabelConv l)); simpl; try contradiction; auto.
specialize (SR l).
makenew' (APsi l).
unfold SpecHeap_join.
makenew' (CHypo (LabelConv l)); simpl; auto.
makenew' (CPsi (LabelConv l)); simpl; try contradiction.
unfold SpecHeap_join.
makenew' (CHypo (LabelConv l)); simpl; try contradiction.
makenew' (CPsi (LabelConv l)); simpl; try contradiction.
auto.

rewrite LL in H2.
specialize SR with (LabelConv' f).
rewrite H6 in SR.
rewrite LL in SR.
rewrite H3 in SR.
contradiction.
Qed.

(******************************************************************)
(* Definition of refinement that keeps the same order of instructions *)


Definition OrderRefinement
   (InstrRel : ASCAP.Process.Op -> CSCAP.Process.Op)
   (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
   (ActionRel : Action AState -> Action CState -> Prop)
   (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label)
   (Adp : ASCAP.Process.Decider_to_Prop) (Cdp : CSCAP.Process.Decider_to_Prop)
   (AMach : ASCAP.Process.Machine) (CMach : CSCAP.Process.Machine) :=
   forall APsi CPsi,
   specrel LabelConv ActionRel APsi CPsi ->
   (forall Ag, exists Cg, ActionRel Ag Cg) /\
   (forall Ai, match AMach Ai with 
               | Some _ => exists Cig, CMach (InstrRel Ai) = Some Cig 
               | None => True 
               end) /\
   (forall Ai Ci Aig Cig Ag Cg,
       InstrRel Ai = Ci -> AMach Ai = Some Aig -> CMach Ci = Some Cig ->
       ActionRel Ag Cg -> ActionGTE Ag Aig -> ActionGTE Cg Cig) /\
   (forall Ag Cg Ag' Cg',
       ActionRel Ag Cg -> ActionRel Ag' Cg' -> ActionGTE Ag Ag' -> ActionGTE Cg Cg') /\
   (forall Ag Cg, 
       ActionRel Ag Cg -> ActionGTE Ag (ActionID AState) -> ActionGTE Cg (ActionID CState)) /\
   (forall p Ag Ag' Cg Cg' Cg'',
       ActionRel Ag Cg -> ActionRel Ag' Cg' -> 
       ActionRel (ActionPlus (Adp p) Ag Ag') Cg'' -> excl Adp p ->
       ActionGTE Cg'' (ActionPlus (Cdp (DeciderRel p)) Cg Cg')) /\
   (forall Ag Ag' Cg Cg' Cg'',
       ActionRel Ag Cg -> ActionRel Ag' Cg' -> ActionRel (ActionChain Ag Ag') Cg'' ->
       ActionGTE Cg'' (ActionChain Cg Cg'))
.


Section procrel.
Variable oprel : ASCAP.Process.Op -> CSCAP.Process.Op.
Variable labelconv : ASCAP.Process.Label -> CSCAP.Process.Label.
Variable DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider.
Variable adp : ASCAP.Process.Decider_to_Prop.
Inductive procrel
    : ASCAP.Process.Process -> CSCAP.Process.Process -> Prop :=
  | procrel_nil : procrel ASCAP.Process.Inil CSCAP.Process.Inil
  | procrel_perf : forall i,
       procrel (ASCAP.Process.Iperf i) (CSCAP.Process.Iperf (oprel i))
  | procrel_call : forall l,
       procrel (ASCAP.Process.Icall l) (CSCAP.Process.Icall (labelconv l))
  | procrel_plus : forall p AI1 AI2 CI1 CI2,
       procrel AI1 CI1 -> procrel AI2 CI2 -> excl adp p ->
       procrel (ASCAP.Process.Iplus p AI1 AI2) (CSCAP.Process.Iplus (DeciderRel p) CI1 CI2)
  | procrel_seq : forall AI1 AI2 CI1 CI2,
       procrel AI1 CI1 -> procrel AI2 CI2 ->
       procrel (ASCAP.Process.Iseq AI1 AI2) (CSCAP.Process.Iseq CI1 CI2)
.
End procrel.

Lemma OrderRefinementValid 
  (LabelConv' : CSCAP.Process.Label -> ASCAP.Process.Label)
  (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label) 
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
  (Adp : ASCAP.Process.Decider_to_Prop)
  (Cdp : CSCAP.Process.Decider_to_Prop)
  (LL : forall l, LabelConv (LabelConv' l) = l) :
  forall ActionRel InstrRel AMach CMach,
   OrderRefinement InstrRel DeciderRel ActionRel LabelConv Adp Cdp AMach CMach ->
   ProcRefinement ActionRel (procrel InstrRel LabelConv DeciderRel Adp) LabelConv Adp Cdp AMach CMach.

Proof.
intros.
red; intros.
generalize CI Cg H2 H1; clear H1 H2 Cg CI.
induction H3;
red in H.

(* nil *)
specialize H with Psi CPsi.
firstorder.
inversion_clear H2.
apply WFProc_weaken with (ActionID CState); auto.
apply H6 with (ActionID AState); auto.

(* perf *)
specialize H with Psi CPsi.
firstorder.
inversion_clear H2.
red in H1.
make (CMach (InstrRel i)).
destruct H2 as [Cig | ].
apply WFProc_weaken with Cig.
apply H5 with i (InstrRel i) g g; auto.
constructor; auto.
specialize H4 with i.
rewrite H1 in H4.
rewrite H10 in H4.
inversion H4.
inversion H2.

(* call *)
specialize H with Psi CPsi.
firstorder.
inversion_clear H2.
red in H1.
red in H0.
specialize H0 with l.
rewrite H1 in H0.
makenew' (CPsi (LabelConv l)); try contradiction.
apply WFProc_weaken with a; auto.
apply H6 with g g; auto.

(* weaken *)
set (y := H); generalize y; clear y; intros.
specialize H with Psi CPsi.
destruct H; [firstorder | ].
specialize H with g'.
inversion H.
apply WFProc_weaken with x; auto.
decompose record H5; clear H5.
apply H8 with g g'; auto.

(* chain *)
set (y := H); generalize y; clear y; intros.
specialize H with Psi CPsi.
destruct H; [firstorder | ].
decompose record H3; clear H3.
inversion_clear H2.
set (xxx := H g').
destruct xxx as [Cg'].
set (xxx := H g'').
destruct xxx as [Cg''].
apply WFProc_weaken with (ActionChain Cg' Cg''); auto.
apply H10 with g' g''; auto.

(* plus *)
set (zzz := H); generalize zzz; clear zzz; intros.
specialize H with Psi CPsi.
destruct H; [firstorder | ].
decompose record H3; clear H3.
inversion_clear H2.
set (xxx := H g').
destruct xxx as [Cg'].
set (xxx := H g'').
destruct xxx as [Cg''].
apply WFProc_weaken with (ActionPlus (Cdp (DeciderRel p)) Cg' Cg''); auto.
apply H8 with g' g''; auto.
Qed.


Lemma OrderRefinementValid'
  (LabelConv' : CSCAP.Process.Label -> ASCAP.Process.Label)
  (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label) 
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
  (Adp : ASCAP.Process.Decider_to_Prop)
  (Cdp : CSCAP.Process.Decider_to_Prop)
  (LL : forall l, LabelConv (LabelConv' l) = l) :
  forall ActionRel InstrRel AMach CMach,
   OrderRefinement InstrRel DeciderRel ActionRel LabelConv Adp Cdp AMach CMach ->
    Refinement (specrel LabelConv ActionRel)
              (coderel LabelConv (procrel InstrRel LabelConv DeciderRel Adp))
              Adp Cdp AMach CMach.
Proof.
intros.
eapply ProcRefinementValid; eauto.
eapply OrderRefinementValid; eauto.
Qed.


(******************************************************************************)
(* The actual REPR refinement definition *)

(* Specification conversion based on REPR *)

Definition reprAction : 
     Action AState -> Action CState := fun g =>
mkAct (fun s => exists bs, repr bs s /\ g.(pre) bs)
      (fun s s' => forall bs, repr bs s -> g.(pre) bs -> 
                (exists bs', repr bs' s' /\ g.(rel) bs bs')).

Definition reprrel : Action AState -> Action CState -> Prop :=
  fun (Ag : Action AState) (Cg : Action CState) => 
       ActionGTE Cg (reprAction Ag) /\  ActionGTE (reprAction Ag) Cg.


(* Properties of REPR-based specification conversion *)

Lemma ActionGTE_repr : forall Ag Ag', 
    ActionGTE Ag Ag' -> ActionGTE (reprAction Ag) (reprAction Ag').
firstorder.
Qed.

Lemma ActionGTE_repr_ID : 
    ActionGTE (reprAction (ActionID AState)) (ActionID CState).
    unfold reprAction; red; intros; intuition; simpl in *; auto.
    subst; firstorder.
Qed.

Lemma ActionGTE_repr_Chain : forall Ag Ag' Cg Cg',
      ActionGTE (reprAction Ag) Cg ->
      ActionGTE (reprAction Ag') Cg' ->
      ActionGTE (reprAction (ActionChain Ag Ag')) (ActionChain Cg Cg').
unfold ActionGTE, reprAction, ActionChain; simpl.
intros; firstorder.
apply H0; firstorder.
specialize (H1 s s'); firstorder.

specialize (H2 x0 s'); firstorder.
specialize (H1 s x0); firstorder.
specialize (H1 bs); firstorder.
specialize (H1 bs); firstorder.
specialize (H7 x1); firstorder.
specialize (H2 x1); firstorder.
Qed.

Lemma ActionGTE_repr_Plus : forall p Ag Ag' Cg Cg'
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
  (Adp : ASCAP.Process.Decider_to_Prop)
  (Cdp : CSCAP.Process.Decider_to_Prop)
  (Decider_repr : forall d sa sc, repr sa sc -> Adp d sa -> Cdp (DeciderRel d) sc)
  (Decider_repr' : forall d sa sc, repr sa sc -> Cdp (DeciderRel d) sc -> Adp d sa)
,
      excl Adp p ->
      ActionGTE (reprAction Ag) Cg ->
      ActionGTE (reprAction Ag') Cg' ->
      ActionGTE (reprAction (ActionPlus (Adp p) Ag Ag')) 
                                         (ActionPlus (Cdp (DeciderRel p)) Cg Cg').
unfold excl, Rrel, ActionGTE, reprAction, ActionChain; simpl.
intros; intuition.

destruct H1; intuition.
left; intuition.
apply H2; intuition.
exists x; intuition.
apply Decider_repr with x; auto.
right; intuition.
apply H0; intuition.
exists x; auto.
apply H7; clear H7.
apply Decider_repr' with s; auto.

specialize (H3 s s').
destruct H3 with bs; intuition.
exists bs; auto.
apply H6.
apply Decider_repr with bs; auto.
exists x; intuition.

specialize (H4 s s').
destruct H4 with bs; intuition.
exists bs; auto.
apply H7.
intros.
apply H10; clear H10.
apply Decider_repr' with s; auto.
exists x; intuition.
Qed.

(* The actual REPR-refinement definition *)

Definition ReprRefinement
   (InstrRel : ASCAP.Process.Op -> CSCAP.Process.Op)
   (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label)
   (AMach : ASCAP.Process.Machine) (CMach : CSCAP.Process.Machine) 
   :=
   forall APsi CPsi,
   specrel LabelConv (reprrel) APsi CPsi ->
   (forall Ai Aig, 
          AMach Ai = Some Aig -> 
          exists Cig, CMach (InstrRel Ai) = Some Cig /\ ActionGTE (reprAction Aig) Cig).

(* Proof the ReprRefinement is a special case of OrderRefinement *)

Lemma ReprRefinementValid
  (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label) 
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
  (Adp : ASCAP.Process.Decider_to_Prop)
  (Cdp : CSCAP.Process.Decider_to_Prop)
  (Decider_repr : forall p sa sc, repr sa sc -> Adp p sa -> Cdp (DeciderRel p) sc)
  (Decider_repr' : forall p sa sc, repr sa sc -> Cdp (DeciderRel p) sc -> Adp p sa)

:
  forall InstrRel AMach CMach,
   ReprRefinement InstrRel LabelConv AMach CMach -> 
   OrderRefinement InstrRel (DeciderRel) (reprrel) LabelConv Adp Cdp AMach CMach.
Proof.
unfold ReprRefinement, OrderRefinement, excl, reprrel;
intros.
specialize (H APsi CPsi); intuition.

exists (reprAction Ag); auto.
makenew' (AMach Ai); auto.
specialize (H1 Ai a); firstorder.

specialize (H1 Ai Aig); intuition.
destruct H4; intuition.
subst.
rewrite H4 in H3; inversion H3; subst.
apply ActionGTE_trans with (reprAction Aig); auto.
apply ActionGTE_trans with (reprAction Ag); auto.
apply ActionGTE_repr; auto.

apply ActionGTE_trans with (reprAction Ag'); auto.
apply ActionGTE_trans with (reprAction Ag); auto.
apply ActionGTE_repr; auto.

apply ActionGTE_trans with (reprAction Ag); auto.
apply ActionGTE_trans with (reprAction (ActionID AState)); auto.
apply ActionGTE_repr; auto.
apply ActionGTE_repr_ID; auto.

apply ActionGTE_trans with (reprAction (ActionPlus (Adp p) Ag Ag')); auto.
apply ActionGTE_repr_Plus; auto.

apply ActionGTE_trans with (reprAction (ActionChain Ag Ag')); auto.
apply ActionGTE_repr_Chain; auto.
Qed.

(* Corollary: ReprRefinement is a Valid Refinement *)

Lemma ReprRefinementValid'
  (LabelConv' : CSCAP.Process.Label -> ASCAP.Process.Label)
  (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label) 
  (LL : forall l, LabelConv (LabelConv' l) = l)  
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
  (Adp : ASCAP.Process.Decider_to_Prop)
  (Cdp : CSCAP.Process.Decider_to_Prop)
  (Decider_repr : forall d sa sc, repr sa sc -> Adp d sa -> Cdp (DeciderRel d) sc)
  (Decider_repr' : forall d sa sc, repr sa sc -> Cdp (DeciderRel d) sc -> Adp d sa)
: 
  forall InstrRel AMach CMach,
   ReprRefinement InstrRel LabelConv AMach CMach -> 
    Refinement (specrel LabelConv reprrel)
              (coderel LabelConv (procrel InstrRel LabelConv DeciderRel Adp))
              Adp Cdp AMach CMach.
intros.
eapply OrderRefinementValid'; auto.
eapply ReprRefinementValid; auto.
Qed.

(*************************************************************)
(* Several simplifications of ReprRefinement Definition *)
(* Simplifications come from using equivalent operations and labels in both machines *)
(* In the paper, the entire system of refinements makes this assumption
    but in the code, we stay more general, then simplify to get similar result
*)

Definition specmake
     (LabelConv' : CSCAP.Process.Label -> ASCAP.Process.Label)
      (APsi : ASCAP.SpecHeap.SpecHeap) : CSCAP.SpecHeap.SpecHeap :=
      fun l => 
      match APsi (LabelConv' l) with
      | Some Ag => Some (reprAction Ag)
      | None => None
      end.

Fixpoint procmake 
  (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label)
  (InstrRel : ASCAP.Process.Op -> CSCAP.Process.Op) 
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
  (i : ASCAP.Process.Process) : CSCAP.Process.Process :=
  match i with
  | ASCAP.Process.Inil => CSCAP.Process.Inil
  | ASCAP.Process.Iperf i => CSCAP.Process.Iperf (InstrRel i)
  | ASCAP.Process.Icall f => CSCAP.Process.Icall (LabelConv f)
  | ASCAP.Process.Iseq I1 I2 => 
        CSCAP.Process.Iseq 
        (procmake LabelConv InstrRel DeciderRel I1)
        (procmake LabelConv InstrRel DeciderRel I2)
  | ASCAP.Process.Iplus p I1 I2 =>
        CSCAP.Process.Iplus (DeciderRel p) 
        (procmake LabelConv InstrRel DeciderRel I1)
        (procmake LabelConv InstrRel DeciderRel I2)
  end.

Definition codemake 
  (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label)
  (LabelConv' : CSCAP.Process.Label -> ASCAP.Process.Label)
  (InstrRel : ASCAP.Process.Op -> CSCAP.Process.Op) 
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
      (AC : ASCAP.CodeHeap) : CSCAP.CodeHeap :=
      fun l => 
      match AC (LabelConv' l) with
      | Some i => Some (procmake LabelConv InstrRel DeciderRel i)
      | None => None
      end.

Lemma specrel_specmake : forall 
  (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label)
  (LabelConv' : CSCAP.Process.Label -> ASCAP.Process.Label)
  Psi,
   (forall l : Label, LabelConv' (LabelConv l) = l) ->
   specrel LabelConv reprrel Psi (specmake LabelConv' Psi).
intros.
unfold specmake.
red; intros; simpl.
rewrite H.
makenew' (Psi l); auto.
red; split; auto.
Qed.


Lemma procrel_procmake : forall 
  (LabelConv : ASCAP.Process.Label -> CSCAP.Process.Label)
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
  (Adp : ASCAP.Process.Decider_to_Prop)
  InstrRel p,
  excl_Proc Adp p ->
procrel InstrRel LabelConv DeciderRel Adp p (procmake LabelConv InstrRel DeciderRel p).
induction p; intros; simpl.

constructor.
constructor.
constructor.
inversion_clear H.
constructor.
apply IHp1; auto.
apply IHp2; auto.
inversion_clear H.
constructor.
apply IHp1; auto.
apply IHp2; auto.
auto.
Qed.

Lemma coderel_codemake : forall LabelConv LabelConv' InstrRel AC
   (lconv_eq : forall l : Label, LabelConv' (LabelConv l) = l)
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
  (Adp : ASCAP.Process.Decider_to_Prop)
,
excl_Code Adp AC ->
coderel LabelConv (procrel InstrRel LabelConv DeciderRel Adp) AC
  (codemake LabelConv LabelConv' InstrRel DeciderRel AC).
intros.
unfold codemake.
red; intros; simpl.
rewrite lconv_eq.
makenew' (AC l); auto.
red in H.
specialize (H l p H1).
apply procrel_procmake; auto.
Qed.


Lemma AutoReprRefinement : forall InstrRel LabelConv LabelConv' AMach CMach
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
  (Adp : ASCAP.Process.Decider_to_Prop)
  (Cdp : CSCAP.Process.Decider_to_Prop)
  (Decider_repr : forall d sa sc, repr sa sc -> Adp d sa -> Cdp (DeciderRel d) sc)
  (Decider_repr' : forall d sa sc, repr sa sc -> Cdp (DeciderRel d) sc -> Adp d sa)
,
   (forall l : Label, LabelConv (LabelConv' l) = l) ->
   (forall l : Label, LabelConv' (LabelConv l) = l) ->
   ReprRefinement InstrRel LabelConv AMach CMach ->
   forall AHypo AC APsi,
   excl_Code Adp AC ->
   ASCAP.WFCode Adp AMach AHypo AC APsi -> 
   CSCAP.WFCode Cdp CMach 
       (specmake LabelConv' AHypo) 
       (codemake LabelConv LabelConv' InstrRel DeciderRel AC)
       (specmake LabelConv' APsi).
intros.
set (x := ReprRefinementValid').
specialize (x LabelConv' LabelConv H DeciderRel Adp Cdp Decider_repr Decider_repr' InstrRel AMach CMach H1).
red in x.
apply x with AC APsi AHypo; auto.
apply specrel_specmake; auto.
apply specrel_specmake; auto.
apply coderel_codemake; auto.
Qed.

(* If we Require that the labels match, we can make it even more automatic *)

Definition EqLabel (l : Label) := l.

Lemma xx1 :  forall l : Label, EqLabel (EqLabel l) = l.
auto.
Qed.

Definition Eq_specmake
      (APsi : ASCAP.SpecHeap.SpecHeap) : CSCAP.SpecHeap.SpecHeap :=
      fun l => 
      match APsi l with
      | Some Ag => Some (reprAction Ag)
      | None => None
      end.

Fixpoint Eq_procmake 
  (InstrRel : ASCAP.Process.Op -> CSCAP.Process.Op) 
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
  (i : ASCAP.Process.Process) 
: CSCAP.Process.Process :=
  match i with
  | ASCAP.Process.Inil => CSCAP.Process.Inil
  | ASCAP.Process.Iperf i => CSCAP.Process.Iperf (InstrRel i)
  | ASCAP.Process.Icall f => CSCAP.Process.Icall f
  | ASCAP.Process.Iseq I1 I2 => 
        CSCAP.Process.Iseq 
        (Eq_procmake InstrRel DeciderRel I1)
        (Eq_procmake InstrRel DeciderRel I2)
  | ASCAP.Process.Iplus p I1 I2 =>
        CSCAP.Process.Iplus (DeciderRel p) 
        (Eq_procmake InstrRel DeciderRel I1)
        (Eq_procmake InstrRel DeciderRel I2)
  end.

Definition Eq_codemake 
  (InstrRel : ASCAP.Process.Op -> CSCAP.Process.Op) 
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
      (AC : ASCAP.CodeHeap) : CSCAP.CodeHeap :=
      fun l => 
      match AC l with
      | Some i => Some (Eq_procmake InstrRel DeciderRel i)
      | None => None
      end.

Lemma specrel_Eq_specmake : forall  Psi,
   specrel EqLabel reprrel Psi (Eq_specmake Psi).
intros.
unfold Eq_specmake.
unfold EqLabel.
red; intros; simpl.
makenew' (Psi l); auto.
red; split; auto.
Qed.

Lemma procrel_Eq_procmake : forall 
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
  (Adp : ASCAP.Process.Decider_to_Prop)
  InstrRel p,
  excl_Proc Adp p ->
procrel InstrRel EqLabel DeciderRel Adp p (Eq_procmake InstrRel DeciderRel p).
induction p; intros; simpl.

constructor.
constructor.
constructor.
inversion_clear H.
constructor.
apply IHp1; auto.
apply IHp2; auto.
inversion_clear H.
constructor.
apply IHp1; auto.
apply IHp2; auto.
auto.
Qed.

Lemma coderel_Eq_codemake : forall InstrRel AC
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
  (Adp : ASCAP.Process.Decider_to_Prop)
,
excl_Code Adp AC ->
coderel EqLabel (procrel InstrRel EqLabel DeciderRel Adp) AC
  (Eq_codemake InstrRel DeciderRel AC).
intros.
unfold Eq_codemake.
unfold EqLabel.
red; intros; simpl.
makenew' (AC l); auto.
red in H.
specialize (H l p H1).
apply procrel_Eq_procmake; auto.
Qed.

Lemma EqAutoReprRefinement {InstrRel AMach CMach} 
  (DeciderRel : ASCAP.Process.Decider -> CSCAP.Process.Decider)
  (Adp : ASCAP.Process.Decider_to_Prop)
  (Cdp : CSCAP.Process.Decider_to_Prop)
  (Decider_repr : forall d sa sc, repr sa sc -> Adp d sa -> Cdp (DeciderRel d) sc)
  (Decider_repr' : forall d sa sc, repr sa sc -> Cdp (DeciderRel d) sc -> Adp d sa)
: 
   ReprRefinement InstrRel EqLabel AMach CMach ->
   forall AHypo AC APsi,
   excl_Code Adp AC ->
   ASCAP.WFCode Adp AMach AHypo AC APsi -> 
   CSCAP.WFCode Cdp CMach 
       (Eq_specmake AHypo) 
       (Eq_codemake InstrRel DeciderRel AC)
       (Eq_specmake APsi).
intros.
set (x := ReprRefinementValid').
specialize (x EqLabel EqLabel xx1 DeciderRel Adp Cdp Decider_repr Decider_repr' InstrRel AMach CMach H).
red in x.
apply x with AC APsi AHypo; auto.
apply specrel_Eq_specmake; auto.
apply specrel_Eq_specmake; auto.
apply coderel_Eq_codemake; auto.
Qed.

End ReprFinement.

End Refinement.