(* mem_alloc.v *)
(* Definition and certification of code of mem_alloc, and its subprocedures *)

Require Import List.
Require Import NArith.
Require Import ZArith.
Require Import Bool.
Require Import Logic.FunctionalExtensionality.

Require Import spec.
Require Import label.

Require Import zbool_assist.
Require Import base_simp.
Require Import Omega.

Require Import c.
Require Import pe.
Require Import mach.mem_sig.

Module mem_alloc.
Import C_PE State Mem Stack Ops SCAP Process.

Require Import sact.
Require Import actsmp.

Opaque PTROOT PE_State.PTROOT PE_State.PMM_data_loc.


(***************************************************)

Definition junk_var : Var := 1%Z.
Definition page_var : Var := 2%Z.
Definition entry_var : Var := 3%Z.
Definition found_var : Var := 4%Z.

Definition mem_alloc_loop_spec :=
{| pre := (fun s =>
    StackCheck isDataFrame s = true /\ 
     StackCheck (getVarCheck entry_var) s = true /\ StackGet (getVar entry_var) s = 0%Z /\
     StackCheck (getVarCheck page_var) s = true /\ StackCheck (getVarCheck found_var) s = true /\ 
     (1 <=  (StackGet (getVar page_var) s) <= NPAGES)%Z /\ 
    MemGet (getTLB) s = true /\ DM s.(memory)) ;

   rel := (fun s s' =>
    if (Zeq_bool (StackGet (getVar found_var) s) 0%Z) then
          (exists page, (StackGet (getVar page_var) s <= page < NPAGES)%Z /\ (MemGet (getPageInfo page) s = false) /\ 
                               (s' = StackAct (setVar page_var page +++ setVar found_var 1%Z) s))
          \/
          (s' = StackAct (setVar page_var NPAGES) s)
    else s' =s)
|}.

Definition mem_alloc_body_spec :=
{| pre := (fun s =>
   StackCheck (getVarCheck found_var) s = true /\ StackGet (getVar found_var) s = 0%Z /\
   StackCheck (getVarCheck entry_var) s = true /\ StackGet (getVar entry_var) s = 0%Z /\
   StackCheck (getVarCheck page_var) s = true /\ StackCheck isDataFrame s = true /\
   (1 <= StackGet (getVar page_var) s < NPAGES)%Z /\
   MemGet (getTLB) s = true /\ DM s.(memory)) ;

    rel := (fun s s' => 
      if MemGet (getPageInfo (StackGet (getVar page_var) s)) s then
      (s' = StackAct (setVar page_var (StackGet (getVar page_var) s + 1)%Z) s)
      else
      (s' = StackAct (setVar found_var 1%Z) s)
      )
|}.

Definition Psi (f : Label) : option (Action State) := 
   match f with
   | mem_alloc_body_label => Some mem_alloc_body_spec
   | mem_alloc_loop_label => Some mem_alloc_loop_spec
   | _ => None
   end.

Notation "a ::: b" := (Iseq a b) (at level 80).
Notation "a ;; b" := (ActionChain a b) (at level 80).


Definition if1 := expr_nomem_binop EQ (expr_nomem_var entry_var) (expr_nomem_z 0%Z).

Require Import IfProp.

Definition PMM_expr (pg : expr) : expr :=
  (expr_binop PLUS (expr_z PMM_data_loc) (expr_binop TIMES pg (expr_z 8%Z))).

Definition mem_alloc_body : Process :=
     Iperf (perf_assign entry_var (expr_mem (PMM_expr (expr_var page_var)))) :::
     Iplus (if1)
        (Iperf (perf_assign found_var (expr_z 1%Z)))
        (Iperf (perf_assign page_var (expr_binop PLUS (expr_var page_var) (expr_z 1%Z)))) :::
     Iperf (perf_assign entry_var (expr_z 0%Z)).

Definition if2 := expr_nomem_binop LOGIC_AND 
                             (expr_nomem_binop LT (expr_nomem_var page_var) (expr_nomem_z NPAGES))
                             (expr_nomem_binop EQ (expr_nomem_var found_var) (expr_nomem_z 0%Z)).

Definition mem_alloc_loop : Process :=
  Iplus (if2)
     (
     Icall mem_alloc_body_label :::
     Icall mem_alloc_loop_label
     )
     Inil.
  

Definition if3 := expr_nomem_binop EQ (expr_nomem_var found_var) (expr_nomem_z 1%Z).

Definition mem_alloc : Process :=
  Iperf (perf_readargs nil) :::
  Iperf (perf_assign page_var (expr_z 1%Z)) :::
  Iperf (perf_assign found_var (expr_z 0%Z)) :::
  Iperf (perf_assign entry_var (expr_z 0%Z)) :::
  Icall mem_alloc_loop_label :::
  Iplus (if3) 
        (
        Iperf (perf_store (PMM_expr (expr_var page_var)) (expr_z 1%Z)) :::
        Iperf (perf_ret (expr_var page_var))
        )
        (Iperf (perf_ret (expr_z 0%Z))).


Definition mem_alloc_body_spec' : Action State :=
     Actions (perf_assign entry_var (expr_mem (PMM_expr (expr_var page_var)))) ;;
     ActionPlus (expr_to_Prop if1)
        (Actions (perf_assign found_var (expr_z 1%Z)))
        (Actions (perf_assign page_var (expr_binop PLUS (expr_var page_var) (expr_z 1%Z))))
        ;;
     Actions (perf_assign entry_var (expr_z 0%Z)).
  

Definition mem_alloc_loop_spec' : Action State :=
  ActionPlus (expr_to_Prop if2)
     (mem_alloc_body_spec ;;
     mem_alloc_loop_spec
     )
     (ActionID State)
.
  
Definition mem_alloc_spec' : Action State :=
  Actions (perf_readargs nil) ;;
  Actions (perf_assign page_var (expr_z 1%Z)) ;;
  Actions (perf_assign found_var (expr_z 0%Z)) ;;
  Actions (perf_assign entry_var (expr_z 0%Z)) ;;
  mem_alloc_loop_spec ;;
  ActionPlus (expr_to_Prop if3) 
        (
        Actions (perf_store (PMM_expr (expr_var page_var)) (expr_z 1%Z)) ;;
        Actions (perf_ret (expr_var page_var)
        ))
        (Actions (perf_ret (expr_z 0%Z))).

Definition WF_mem_alloc_body' : WFProc expr_to_Prop Mc_Machine Psi mem_alloc_body_spec' mem_alloc_body.
unfold mem_alloc_body_spec'; unfold mem_alloc_body; auto 999.
Qed.

Definition WF_mem_alloc_loop' : WFProc expr_to_Prop Mc_Machine Psi mem_alloc_loop_spec' mem_alloc_loop.
unfold mem_alloc_loop_spec'; unfold mem_alloc_loop; auto 999.
Qed.

Definition WF_mem_alloc' : WFProc expr_to_Prop Mc_Machine Psi mem_alloc_spec' mem_alloc.
unfold mem_alloc_spec'; unfold mem_alloc; auto 999.
Qed.

Definition WF_mem_alloc : WFProc expr_to_Prop Mc_Machine Psi mem_sig.mem_alloc_spec mem_alloc.
apply WFProc_weaken with mem_alloc_spec'; [|apply WF_mem_alloc'].
unfold mem_sig.mem_alloc_spec, mem_alloc_spec', mem_alloc_loop_spec, if3, expr_to_Prop.

C_auf.
sact_structured.
sact_auf.
unfold StackAct, MemAct, MemCheck, StackCheck, StackGet, MemGet in *; simpl in *.
split; intros; simpl in *; repeat (autorewrite with c stack mem in *; simpl in *).

unfold StackAct, MemAct, MemCheck, StackCheck, StackGet, MemGet in *; simpl in *.
split; intros; repeat (autorewrite with base c mem stack in *; simpl in *).
repeat autoclean.
repeat (split; autosolve);
try (intros; subst; repeat autoclean; simpl in *; autosolve).

unfold getArgsCheck, getVarCheck, setMemCheck, isFrame in *; simpl in *.
destruct s; simpl in *.
destruct stack0; simpl in *; autosolve.
destruct f; simpl in *; autosolve.
destruct l; simpl in *; autosolve.
repeat autoclean; simpl in *.
destruct H2; repeat autoclean; simpl in *.
left.
repeat split; autosolve.

unfold ValidVA; simpl in *.
unfold getTLB in H; rewrite H; simpl.
unfold getPageInfo in H2.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
red in H1.
rewrite H1; autosolve.
rewrite <- Addr_parts; autosolve.
rewrite PE_State.PMMDom_Valid; autosolve.
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
apply ValidPA_ValidPg.
rewrite PE_State.PMMDom_Valid; autosolve.
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

intros; subst; simpl in *; auto.
right; split; autosolve.

(* rel *)
unfold StackAct, MemAct, MemCheck, StackCheck, StackGet, MemGet in *; simpl in *.
repeat autoclean.
unfold getVar, getArgsCheck, setMem in *; simpl in *.
destruct s; simpl in *.
destruct stack0; simpl in *; autosolve.
destruct f; simpl in *; autosolve.
destruct l; simpl in *; autosolve.
unfold mem_sig.allocpg; simpl in *.

destruct H6.
(* exists page *)
left.
destruct H0 as [page].
exists page.
repeat autoclean; simpl in *.
destruct H4; simpl in *; try split; autosolve.
repeat autoclean; simpl in *.
unfold getPageInfo, setPageInfo in *; simpl in *.
Zeq_pull'; Zeq_norm; simpl in *; repeat (split; autosolve).
f_equal.
apply functional_extensionality; intros.
rewrite H3.
rewrite <- Addr_parts; autosolve.
apply ValidPA_ValidPg.
apply PE_State.PMMDom_Valid.
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
(* opposite case *)
simpl in *.
right; subst; simpl in *.
rewrite <- H5; simpl in *; autosolve.
Qed.


Definition WF_mem_alloc_body : WFProc expr_to_Prop Mc_Machine Psi mem_alloc_body_spec mem_alloc_body.
apply WFProc_weaken with mem_alloc_body_spec'; [|apply WF_mem_alloc_body'].
unfold mem_alloc_body_spec; unfold mem_alloc_body_spec', if1, expr_to_Prop.

C_auf.
sact_structured.
sact_auf.
unfold StackAct, MemAct, MemCheck, StackCheck, StackGet, MemGet, PE_State.getMemCheck, ValidVA in *; simpl in *.
split; intros; simpl in *; repeat (autorewrite with c stack mem in *; simpl in *).

unfold StackAct, MemAct, MemCheck, StackCheck, StackGet, MemGet in *; simpl in *.
repeat autoclean.
unfold getVarCheck, setVarCheck, getVar, getTLB, isDataFrame, 
        C_AS.State.Stack.isDataFrame, C_AS.State.Stack.getVarCheck in *;
        simpl in *.
destruct s; simpl in *.
destruct stack0; simpl in *; autosolve.
destruct f; simpl in *; autosolve.
makenew' (d found_var); autosolve.
makenew' (d entry_var); autosolve.
makenew' (d page_var); autosolve.
subst.
repeat autoclean.
rewrite H6; simpl.
rewrite H7.
rewrite <- Addr_parts.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite PE_State.PMMDom_Valid; autosolve.
repeat (split; autosolve).

intros.
subst; simpl.
rewrite H12.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
left; repeat split; autosolve.
right; repeat split; autosolve.

intros; simpl in *.
repeat autoclean; simpl in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite <- H0; simpl in *; try split; autosolve.
rewrite <- H1; simpl in *; try split; autosolve.

unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

apply ValidPA_ValidPg.
rewrite PE_State.PMMDom_Valid; autosolve.
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

(* rel *)
unfold StackAct, MemAct, MemCheck, StackCheck, StackGet, MemGet in *; simpl in *.
repeat autoclean.
unfold getVarCheck, setVarCheck, getVar, getTLB, isDataFrame, getPageInfo, setPageInfo,
        C_AS.State.Stack.isDataFrame, C_AS.State.Stack.getVarCheck, PE_State.getMem in *;
        simpl in *.
destruct s; simpl in *.
destruct stack0; simpl in *; autosolve.
destruct f; simpl in *; autosolve.
makenew' (d found_var); autosolve.
makenew' (d entry_var); autosolve.
makenew' (d page_var); autosolve.
subst.
simpl in *.
rewrite H8 in *.
rewrite <- Addr_parts in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite <- H11; simpl in *; repeat (split; autosolve).
repeat f_equal.
apply functional_extensionality; intros.
unfold Var_beq.
Zeq_pull'; Zeq_norm; simpl in *; subst; autosolve.

rewrite <- H12; simpl in *; repeat (split; autosolve).
repeat f_equal.
apply functional_extensionality; intros.
unfold Var_beq.
Zeq_pull'; Zeq_norm; simpl in *; subst; autosolve.
red; intros; autosolve.

apply ValidPA_ValidPg.
apply PE_State.PMMDom_Valid.
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

apply ValidPA_ValidPg.
apply PE_State.PMMDom_Valid.
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

apply ValidPA_ValidPg.
apply PE_State.PMMDom_Valid.
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

Qed.

Definition WF_mem_alloc_loop : WFProc expr_to_Prop Mc_Machine Psi mem_alloc_loop_spec mem_alloc_loop.
apply WFProc_weaken with mem_alloc_loop_spec'; [|apply WF_mem_alloc_loop'].
unfold mem_alloc_loop_spec, mem_alloc_loop_spec'.
unfold mem_alloc_body_spec, mem_alloc_loop_spec.
unfold if2, expr_to_Prop.

C_auf.
sact_structured.
sact_auf.
unfold StackAct, MemAct, MemCheck, StackCheck, StackGet, MemGet in *; simpl in *.
split; intros; simpl in *; repeat (autorewrite with c stack mem in *; simpl in *).

unfold StackAct, MemAct, MemCheck, StackCheck, StackGet, MemGet in *; simpl in *.
repeat autoclean.
unfold getVarCheck, setVarCheck, getVar, getTLB, isDataFrame, getPageInfo, setPageInfo,
        C_AS.State.Stack.isDataFrame, C_AS.State.Stack.getVarCheck in *;
        simpl in *.
destruct s; simpl in *.
destruct stack0; simpl in *; autosolve.
destruct f; simpl in *; autosolve.
makenew' (d found_var); autosolve.
makenew' (d entry_var); autosolve.
makenew' (d page_var); autosolve.
subst.
simpl in *.
repeat autoclean.
Zeq_pull'; Zeq_norm; simpl in *; subst; simpl in *; autosolve.
left.
repeat (split; autosolve); subst; simpl; autosolve.
rewrite H10; auto.
rewrite H10; auto.
rewrite H11; auto.
rewrite H11; auto.
rewrite H11; auto.

right; repeat (split; autosolve).
right; repeat (split; autosolve).
right; repeat (split; autosolve).

left.
repeat (split; autosolve); subst; simpl; autosolve.
rewrite H10; auto.
rewrite H10; auto.
rewrite H9; auto.

right; repeat (split; autosolve).
right; repeat (split; autosolve).
right; repeat (split; autosolve).

(* rel *)

unfold StackAct, MemAct, MemCheck, StackCheck, StackGet, MemGet in *; simpl in *.
repeat autoclean.
unfold getVarCheck, setVarCheck, getVar, getTLB, isDataFrame, getPageInfo, setPageInfo,
        C_AS.State.Stack.isDataFrame, C_AS.State.Stack.getVarCheck in *;
        simpl in *.
destruct s; simpl in *.
destruct stack0; simpl in *; autosolve.
destruct f; simpl in *; autosolve.
makenew' (d found_var); autosolve.
makenew' (d entry_var); autosolve.
makenew' (d page_var); autosolve.
repeat autoclean.

Zeq_pull'; Zeq_norm; simpl in *; repeat autoclean; autosolve.
destruct H0; try (split; autosolve); repeat autoclean; subst; simpl in *; subst; simpl in *.
left.
exists z1.
rewrite H3.
rewrite Zeq_bool_refl.
repeat (split; autosolve).
repeat f_equal.
apply functional_extensionality; intros.
unfold Var_beq.
Zeq_pull'; Zeq_norm; simpl in *; subst; simpl in *; autosolve.

right.
rewrite H1.
repeat f_equal.
apply functional_extensionality; intros.
unfold Var_beq.
Zeq_pull'; Zeq_norm; simpl in *; subst; simpl in *; autosolve.
assert (z1 = NPAGES); autosolve.
subst; autosolve.
red; intros; autosolve.

destruct H0; try (split; autosolve); repeat autoclean; subst; simpl in *; subst; simpl in *.
rewrite H11 in H0.
rewrite Zeq_bool_refl in H0.
destruct H0; repeat autoclean.
left.
exists x.
repeat (split; autosolve).
repeat f_equal.
apply functional_extensionality; intros.
unfold Var_beq.
Zeq_pull'; Zeq_norm; simpl in *; subst; simpl in *; autosolve.

clear H1.
right.
repeat f_equal.
apply functional_extensionality; intros.
unfold Var_beq.
Zeq_pull'; Zeq_norm; simpl in *; subst; simpl in *; autosolve.

right.
rewrite H1; try (red; intros; autosolve).
repeat f_equal.
apply functional_extensionality; intros.
unfold Var_beq.
Zeq_pull'; Zeq_norm; simpl in *; subst; simpl in *; autosolve.
assert (z1 = NPAGES); autosolve.
subst; autosolve.

rewrite H1; try (red; intros; autosolve); autosolve.
rewrite H1; try (red; intros; autosolve); autosolve.
rewrite H1; try (red; intros; autosolve); autosolve.
rewrite H1; try (red; intros; autosolve); autosolve.
Qed.

End mem_alloc.
